#!/usr/bin/env python
"""
_LoadRunning_

MySQL implementation for loading bl_runjob records for active jobs
"""


from WMCore.Database.DBFormatter import DBFormatter

class LoadRunning(DBFormatter):
    """
    _LoadRunning_

    Load all bl_runjob that are active

    """
    sql = """SELECT rj.wmbs_id jobid, rj.grid_id gridid, rj.bulk_id bulkid,
               st.name status, rj.retry_count retry_count, rj.id id,
               rj.status_time status_time, wu.cert_dn AS userdn,
               wu.group_name AS usergroup, wu.role_name AS userrole,
               wj.cache_dir AS cache_dir
             FROM bl_runjob rj
             INNER JOIN bl_status st ON rj.sched_status = st.id
             LEFT OUTER JOIN wmbs_users wu ON wu.id = rj.user_id
             INNER JOIN wmbs_job wj ON wj.id = rj.wmbs_id
             WHERE rj.status = 1
             """

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        """
        result = self.dbi.processData(self.sql, binds = {}, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_SetOutcomeBulk_

MySQL implementation of Jobs.SetOutcomeBulk
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class SetOutcomeBulk(DBFormatter):
    sql = """UPDATE wmbs_job SET outcome = :outcome
               WHERE id = :jobid"""

    def execute(self, binds, conn = None, transaction = False):
        """
        _execute_

        Expect a list of binds of the type:
        {'jobid': ID, 'outcome': outcome}
        """
        for bind in binds:
            if bind['outcome'] == 'success' \
                   or bind['outcome'] == '1' \
                   or bind['outcome'] == 1:
                bind['outcome'] = 1
            else:
                bind['outcome'] = 0

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_Delete_

Oracle implementation of JobGroup.Delete

"""
__all__ = []



from WMCore.WMBS.MySQL.JobGroup.Delete import Delete as DeleteJobGroupMySQL

class Delete(DeleteJobGroupMySQL):
    sql = DeleteJobGroupMySQL.sql



#!/usr/bin/env python
"""
_RandomSeeder_

Simple Random seed generator

"""
import random
from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface



class BasicRandom(GeneratorInterface):
    """
    Generator that generates a set of random numbers for a job
    and inserts them into the job baggage

    """
    def __init__(self, **options):
        GeneratorInterface.__init__(self, **options)
        self.instance = random.SystemRandom()
        self.maxInt   = options.get("maxint", 900000000)
        self.seeds    = options.get("seed_list", [])


    def __call__(self, wmbsJob):
        baggage = wmbsJob.getBaggage()
        for x in self.options['seed_list']:
            wmbsJob.addBaggageParameter(
                "RandomSeeder.%s" %(x),
                self.instance.randint(1, self.maxInt)
                )



#!/usr/bin/env python
"""
_IsAllWorkflowsCompleted_

MySQL implementation of DBSBuffer.IsAllWorkflowsCompleted

"""

from __future__ import print_function, division
from WMCore.Database.DBFormatter import DBFormatter


class IsAllWorkflowsCompleted(DBFormatter):
    """
    _IsAllWorkflowsCompleted_

    check whether all workflows are completed
    """
    sql = """SELECT count(*) as count FROM dbsbuffer_workflow WHERE completed=0"""

    def execute(self, conn=None, transaction=False):
        """
        returns if there is no commpleted == 0 tasks which means all the workflows are in completed status
        """
        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)

        flag = True if self.format(result)[0][0] == 0 else False
        return flag



#!/usr/bin/env python
"""
_UpdateJobs_

Oracle implementation for updating jobs
"""


from WMCore.BossAir.MySQL.UpdateJobs import UpdateJobs as MySQLUpdateJobs

class UpdateJobs(MySQLUpdateJobs):
    """
    _UpdateJobs_

    Update jobs with new values
    """



#!/usr/bin/env python
"""
The actual error handler algorithm

The current ErrorHandler will either exhaust jobs based on the retry_count, or move jobs
onward into their necessary cooloff state.  The number of retries a job is allowed is
based upon:

config.ErrorHandler.maxRetries

However, it can also be used to handle jobs based on properties in the FWJR.
In order to engage any of this behavior you have to set the config option:
config.ErrorHandler.readFWJR = True

It will then take three arguments.

config.ErrorHandler.maxFailTime:  This should be a time in seconds after which, if the job took that long to fail,
it should be exhausted immediately.  It defaults to 24 hours.

config.ErrorHandler.passExitCodes:  This should be a list of exitCodes that you want to cause the job to move
immediately to the 'created' state, skipping cooloff.  It defaults to [].

Note that exitCodesNoRetry has precedence over passExitCodes.
"""
from future import standard_library
standard_library.install_aliases()

import logging
import os.path
import threading
from http.client import HTTPException
from Utils.Timers import timeFunction
from Utils.IteratorTools import grouper
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.DAOFactory import DAOFactory
from WMCore.Database.CouchUtils import CouchConnectionError
from WMCore.FwkJobReport.Report import Report
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.WMBS.Job import Job
from WMCore.WMException import WMException
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Services.ReqMgrAux.ReqMgrAux import ReqMgrAux


class ErrorHandlerException(WMException):
    """
    The Exception class for the ErrorHandlerPoller

    """
    pass


class ErrorHandlerPoller(BaseWorkerThread):
    """
    Polls for Error Conditions, handles them
    """

    def __init__(self, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.config = config

        myThread = threading.currentThread()

        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.changeState = ChangeState(self.config)

        if hasattr(self.config, "Tier0Feeder"):
            self.reqAuxDB = None
            self.maxRetries = self.config.ErrorHandler.maxRetries
        else:
            cacheduration = getattr(self.config.General, "ReqMgrAuxCacheDuration", 5 / 60)  # 5 minutes
            self.reqAuxDB = ReqMgrAux(self.config.General.ReqMgr2ServiceURL, httpDict={'cacheduration': cacheduration})

        self.exitCodesNoRetry = []
        self.maxProcessSize = getattr(self.config.ErrorHandler, 'maxProcessSize', 250)
        self.maxFailTime = getattr(self.config.ErrorHandler, 'maxFailTime', 32 * 3600)
        self.readFWJR = getattr(self.config.ErrorHandler, 'readFWJR', False)
        self.passCodes = getattr(self.config.ErrorHandler, 'passExitCodes', [])

        self.getJobs = self.daoFactory(classname="Jobs.GetAllJobs")
        self.idLoad = self.daoFactory(classname="Jobs.LoadFromIDWithType")
        self.loadAction = self.daoFactory(classname="Jobs.LoadForErrorHandler")

        self.dataCollection = DataCollectionService(url=config.ACDC.couchurl,
                                                    database=config.ACDC.database)

        self.setupComponentParam()

        return

    def setupComponentParam(self):
        """
        Initialize (and update every cycle) some of the component's
        parameters, according to the agent type (T0/Production) and agent config.
        :return: True if the setup was completed, False if any non-critical problem happened
        """
        if self.reqAuxDB:
            agentConfig = self.reqAuxDB.getWMAgentConfig(self.config.Agent.hostName)
            if not agentConfig:
                return False

            self.exitCodesNoRetry = agentConfig.get("NoRetryExitCodes", [])
            if agentConfig.get("UserDrainMode") and agentConfig.get("SpeedDrainMode") \
                and agentConfig.get("SpeedDrainConfig")['NoJobRetries']['Enabled']:
                logging.info("Agent is in speed drain mode. Not retrying any jobs.")
                self.maxRetries = 0
            else:
                self.maxRetries = agentConfig.get("MaxRetries")

        if not isinstance(self.maxRetries, dict):
            self.maxRetries = {'default': self.maxRetries}
        if 'default' not in self.maxRetries:
            raise ErrorHandlerException('Max retries for the default job type must be specified')
        return True

    def setup(self, parameters=None):
        """
        Load DB objects required for queries
        """
        # For now, does nothing
        return

    def terminate(self, params):
        """
        _terminate_

        Do one pass, then commit suicide
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)

    def exhaustJobs(self, jobList):
        """
        _exhaustJobs_

        Actually do the jobs exhaustion
        """

        # Remove all the files in the exhausted jobs.
        logging.debug("About to fail input files for exhausted jobs")
        for job in jobList:
            job.failInputFiles()

        # Do not build ACDC for utilitarian job types
        acdcJobList = [job for job in jobList if job['type'] not in ['LogCollect', 'Cleanup']]

        self.handleACDC(acdcJobList)

        self.changeState.propagate(jobList, 'exhausted', 'retrydone')

        return

    def processRetries(self, jobList, state):
        """
        _processRetries_

        Actually do the retries
        """
        logging.info("Processing retries for %d failed jobs of type %sfailed", len(jobList), state)
        retrydoneJobs = []
        cooloffJobs = []
        passJobs = []

        if not isinstance(self.maxRetries, dict):
            self.maxRetries = {'default': self.maxRetries}
        if 'default' not in self.maxRetries:
            raise ErrorHandlerException('Max retries for the default job type must be specified')

        # Retries < max retry count
        for job in jobList:
            allowedRetries = self.maxRetries.get(job['type'], self.maxRetries['default'])
            # Retries < allowed max retry count
            if job['retry_count'] < allowedRetries and state != 'create':
                cooloffJobs.append(job)
            # Check if Retries >= allowed max retry count
            elif job['retry_count'] >= allowedRetries or state == 'create':
                retrydoneJobs.append(job)
                msg = "Stopping retries for job %d" % job['id']
                logging.debug(msg)
                logging.debug("JobInfo: %s", job)

        if self.readFWJR:
            # Then we have to check each FWJR for exit status
            cooloffJobs, passJobs, retrydoneFWJRJobs = self.readFWJRForErrors(cooloffJobs)
            retrydoneJobs.extend(retrydoneFWJRJobs)

        # Now to actually do something.
        logging.debug("About to propagate jobs")
        if len(retrydoneJobs) > 0:
            self.changeState.propagate(retrydoneJobs, 'retrydone',
                                       '%sfailed' % state, updatesummary=True)
        if len(cooloffJobs) > 0:
            self.changeState.propagate(cooloffJobs, '%scooloff' % state,
                                       '%sfailed' % state, updatesummary=True)
        if len(passJobs) > 0:
            # Overwrite the transition states and move directly to created
            self.changeState.propagate(passJobs, 'created', 'new')

        return

    def handleACDC(self, jobList):
        """
        _handleACDC_

        Do the ACDC creation and hope it works
        """
        idList = [x['id'] for x in jobList]
        logging.info("Starting to build ACDC with %i jobs", len(idList))
        logging.info("This operation will take some time...")
        loadList = self.loadJobsFromListFull(idList)
        for job in loadList:
            job.getMask()
        self.dataCollection.failedJobs(loadList)
        return

    def readFWJRForErrors(self, jobList):
        """
        _readFWJRForErrors_

        Check the FWJRs of the failed jobs
        and determine those that can be retried
        and which must be retried without going through cooloff.
        Returns a triplet with cooloff, passed and exhausted jobs.
        """
        cooloffJobs = []
        passJobs = []
        exhaustJobs = []

        for job in jobList:
            report = Report()
            reportPath = job['fwjr_path']
            if reportPath is None:
                logging.error("No FWJR in job %i, ErrorHandler can't process it.\n Passing it to cooloff.", job['id'])
                cooloffJobs.append(job)
                continue
            if not os.path.isfile(reportPath):
                logging.error(
                    "Failed to find FWJR for job %i in location %s.\n Passing it to cooloff.", job['id'], reportPath)
                cooloffJobs.append(job)
                continue
            try:
                report.load(reportPath)
                # First let's check the time conditions
                times = report.getFirstStartLastStop()
                startTime = None
                stopTime = None
                if times is not None:
                    startTime = times['startTime']
                    stopTime = times['stopTime']

                # correct the location if the original location is different from recorded in wmbs
                # WARNING: we are not updating job location in wmbs only updating in couchdb by doing this.
                # If location in wmbs needs to be updated, it should happen in JobAccountant.
                locationFromFWJR = report.getSiteName()
                if locationFromFWJR:
                    job["location"] = locationFromFWJR
                    job["site_cms_name"] = locationFromFWJR

                if startTime is None or stopTime is None:
                    # We have no information to make a decision, keep going.
                    logging.debug("No start, stop times for steps for job %i", job['id'])
                elif stopTime - startTime > self.maxFailTime:
                    msg = "Job %i exhausted after running on node for %i seconds" % (job['id'], stopTime - startTime)
                    logging.debug(msg)
                    exhaustJobs.append(job)
                    continue

                if len([x for x in report.getExitCodes() if x in self.exitCodesNoRetry]):
                    msg = "Job %i exhausted due to a bad exit code (%s)" % (job['id'], str(report.getExitCodes()))
                    logging.debug(msg)
                    exhaustJobs.append(job)
                    continue

                if len([x for x in report.getExitCodes() if x in self.passCodes]):
                    msg = "Job %i restarted immediately due to an exit code (%s)" % (job['id'],
                                                                                     str(report.getExitCodes()))
                    logging.debug(msg)
                    passJobs.append(job)
                    continue

                cooloffJobs.append(job)

            except Exception as ex:
                logging.warning("Exception while trying to check jobs for failures!")
                logging.warning(str(ex))
                logging.warning("Ignoring and sending job to cooloff")
                cooloffJobs.append(job)

        return cooloffJobs, passJobs, exhaustJobs

    def handleRetryDoneJobs(self, jobList):
        """
        _handleRetryDoneJobs_

        """
        myThread = threading.currentThread()
        logging.info("About to process %d retry done jobs", len(jobList))
        myThread.transaction.begin()
        self.exhaustJobs(jobList)
        myThread.transaction.commit()

        return

    def handleFailedJobs(self, jobList, state):
        """
        _handleFailedJobs_

        """
        myThread = threading.currentThread()
        logging.info("About to process %d failures", len(jobList))
        myThread.transaction.begin()
        self.processRetries(jobList, state)
        myThread.transaction.commit()

        return

    def handleErrors(self):
        """
        Queries DB for all watched filesets, if matching filesets become
        available, create the subscriptions
        """
        # Run over created, submitted and executed job failures
        failure_states = ['create', 'submit', 'job']
        for state in failure_states:
            idList = self.getJobs.execute(state="%sfailed" % state)
            logging.info("Found %d failed jobs in state %sfailed", len(idList), state)
            for jobSlice in grouper(idList, self.maxProcessSize):
                jobList = self.loadJobsFromList(jobSlice)
                self.handleFailedJobs(jobList, state)

        # Run over jobs done with retries
        idList = self.getJobs.execute(state='retrydone')
        logging.info("Found %d jobs done with all retries", len(idList))
        for jobSlice in grouper(idList, self.maxProcessSize):
            jobList = self.loadJobsFromList(jobSlice)
            self.handleRetryDoneJobs(jobList)

        return

    def loadJobsFromList(self, idList):
        """
        _loadJobsFromList_

        Load jobs in bulk
        """
        binds = []
        for jobID in idList:
            binds.append({"jobid": jobID})
        results = self.idLoad.execute(jobID=binds)

        # You have to have a list
        if isinstance(results, dict):
            results = [results]

        listOfJobs = []
        for entry in results:
            # One job per entry
            tmpJob = Job(id=entry['id'])
            tmpJob.update(entry)
            listOfJobs.append(tmpJob)

        return listOfJobs

    def loadJobsFromListFull(self, idList):
        """
        _loadJobsFromList_

        Load jobs in bulk.
        Include the full metadata.
        """

        binds = []
        for jobID in idList:
            binds.append({"jobid": jobID})

        results = self.loadAction.execute(jobID=binds)

        # You have to have a list
        if isinstance(results, dict):
            results = [results]

        listOfJobs = []
        for entry in results:
            # One job per entry
            tmpJob = Job(id=entry['id'])
            tmpJob.update(entry)
            listOfJobs.append(tmpJob)

        return listOfJobs

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Performs the handleErrors method, looking for each type of failure
        And deal with it as desired.
        """
        logging.debug("Running error handling algorithm")
        if not self.setupComponentParam():
            msg = "Component failed to retrieve agent configuration from central ReqMgr Aux DB."
            msg += " Skipping this cycle."
            logging.error(msg)
            return

        try:
            myThread = threading.currentThread()
            self.handleErrors()
        except (CouchConnectionError, HTTPException) as ex:
            if getattr(myThread, 'transaction', None) is not None:
                myThread.transaction.rollback()
            msg = "Caught CouchConnectionError/HTTPException exception in ErrorHandler. "
            msg += "Transactions postponed until the next polling cycle\n"
            msg += str(ex)
            logging.error(msg)
        except Exception as ex:
            if getattr(myThread, 'transaction', None) is not None:
                myThread.transaction.rollback()
            msg = "Caught unexpected exception in ErrorHandler:\n"
            msg += str(ex)
            logging.exception(msg)
            raise ErrorHandlerException(msg)



#!/usr/bin/env python
"""
_New_

MySQL implementation of Locations.New
"""

from WMCore.Database.DBFormatter import DBFormatter


class New(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_location (site_name, ce_name,
                                               pending_slots, running_slots,
                                               plugin, cms_name, state, state_time)
                      SELECT
                      :location AS site_name, :cename AS ce_name,
                      :pending_slots AS pending_slots,
                      :running_slots AS running_slots,
                      :plugin AS plugin,
                      :cmsname AS cms_name,
                      (SELECT id FROM wmbs_location_state WHERE name = 'Normal') AS state,
                      :state_time AS state_time"""

    pnnSQL = "INSERT IGNORE INTO wmbs_pnns (pnn) VALUES (:pnn)"

    mapSQL = """INSERT IGNORE INTO wmbs_location_pnns (location, pnn)
                 SELECT wl.id, wpnn.id FROM wmbs_location wl, wmbs_pnns wpnn
                  WHERE wl.site_name = :location AND wpnn.pnn = :pnn"""

    def execute(self, siteName, cmsName=None, ceName=None, pnn="None",
                runningSlots=0, pendingSlots=0, plugin=None, stateTime=None,
                conn=None, transaction=False):
        """
        _execute_

        Now with 100% more plugin support
        """
        binds = {"location": siteName, "cmsname": cmsName, "cename": ceName,
                 "running_slots": runningSlots, "pending_slots": pendingSlots,
                 "plugin": plugin, "state_time": stateTime}
        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)

        binds = {'pnn': pnn}
        self.dbi.processData(self.pnnSQL, binds, conn=conn,
                             transaction=transaction)

        binds = {'location': siteName, 'pnn': pnn}
        self.dbi.processData(self.mapSQL, binds, conn=conn,
                             transaction=transaction)
        return



"""
MySQL implementation of File.GetByLFN
"""
from WMCore.Database.DBFormatter import DBFormatter

class GetByLFN(DBFormatter):
    sql = """SELECT id, lfn, filesize, events, first_event, merged
             FROM wmbs_file_details WHERE lfn = :lfn"""

    def formatDict(self, result):
        """
        _formatDict_

        Cast the integer attributes of the file object to integers as
        formatDict() will turn everything into a string.
        """
        formattedResult = DBFormatter.formatDict(self, result)[0]
        formattedResult["id"] = int(formattedResult["id"])
        formattedResult["merged"] = bool(int(formattedResult["merged"]))

        if formattedResult["events"] != None:
            formattedResult["events"] = int(formattedResult["events"])
        if formattedResult["first_event"] != None:
            formattedResult["first_event"] = int(formattedResult["first_event"])

        if "size" in formattedResult:
            formattedResult["size"] = formattedResult["size"]
        else:
            # The size column is named "filesize" in Oracle as size is
            # as reserved word.  We'll handle this here to make things
            # easier in the Oracle version of this object.
            formattedResult["size"] = formattedResult["filesize"]
            del formattedResult["filesize"]

        if formattedResult["size"] != None:
            formattedResult["size"] = int(formattedResult["size"])

        return formattedResult

    def execute(self, lfn = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"lfn": lfn},
                         conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_LoadFromID_

Oracle implementation of Workflow.LoadFromID

"""
__all__ = []



from WMCore.WMBS.MySQL.Workflow.LoadFromID import LoadFromID as LoadWorkflowMySQL

class LoadFromID(LoadWorkflowMySQL):
    sql = LoadWorkflowMySQL.sql



#!/usr/bin/env python
"""
LogDBBackend

Interface to LogDB persistent storage
"""

from builtins import object, bytes

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytes, encodeUnicodeToBytesConditional

# system modules
import datetime
import hashlib
import time

# WMCore modules
from WMCore.Database.CMSCouch import CouchServer, CouchNotFoundError
from WMCore.Services.LogDB.LogDBExceptions import LogDBError

# define full list of supported LogDB types
LOGDB_MSG_TYPES = ['info', 'error', 'warning', 'comment']

def gen_hash(key):
    "Generate hash for given key"
    key = encodeUnicodeToBytes(key)  # if key is not unicode, then it is not changed
    if not isinstance(key, bytes):
        raise NotImplementedError
    keyhash = hashlib.md5()
    keyhash.update(encodeUnicodeToBytesConditional(key, condition=PY3))
    return keyhash.hexdigest()

def tstamp():
    "Return timestamp with microseconds"
    now = datetime.datetime.now()
    base = str(time.mktime(now.timetuple())).split('.')[0]
    ctime = '%s.%s' % (base, now.microsecond)
    return float(ctime)

def clean_entry(doc):
    """Clean document from CouchDB attributes"""
    for attr in ['_rev', '_id']:
        if  attr in doc:
            del doc[attr]
    return doc

class LogDBBackend(object):
    """
    Represents persistent storage for LogDB
    """
    def __init__(self, db_url, db_name, identifier, thread_name, **kwds):
        self.db_url = db_url
        self.server = CouchServer(db_url)
        self.db_name = db_name
        self.dbid = identifier
        self.thread_name = thread_name
        self.agent = kwds.get('agent', 0)
        self.db = self.server.connectDatabase(db_name, create=False)
        self.design = 'LogDB' # name of design document
        self.view = 'requests' # name of view to look-up requests
        self.tsview = 'tstamp' # name of tsview to look-up requests
        self.threadview = 'logByRequestAndThread'
        self.requestview = 'logByRequest'

    def deleteDatabase(self):
        """Delete back-end database"""
        if  self.db_name in self.server.listDatabases():
            self.server.deleteDatabase(self.db_name)

    def check(self, request, mtype=None):
        """Check that given request name is valid"""
        # TODO: we may add some logic to check request name, etc.
        if  not request:
            raise LogDBError("Request name is empty")
        if  mtype and mtype not in LOGDB_MSG_TYPES:
            raise LogDBError("Unsupported message type: '%s', supported types %s" \
                    % (mtype, LOGDB_MSG_TYPES))

    def docid(self, request, mtype):
        """Generate doc id, we use double dash to avoid dashes from thread names"""
        return gen_hash('--'.join((request, self.dbid, self.thread_name, mtype)))

    def prefix(self, mtype):
        """Generate agent specific prefix for given message type"""
        if  self.agent:
            # we add prefix for agent messages, all others will not have this index
            mtype = 'agent-%s' % mtype
        return mtype

    def agent_update(self, request, msg='', mtype="info"):
        """Update agent info in LogDB for given request"""
        self.check(request, mtype)
        mtype = self.prefix(mtype)
        rec = {"ts":tstamp(), "msg":msg}
        doc = {"_id": self.docid(request, mtype), "messages": [rec],
                "request":request, "identifier":self.dbid,
                "thr":self.thread_name, "type":mtype}
        try:
            exist_doc = self.db.document(doc["_id"])
            doc["_rev"] = exist_doc["_rev"]
        except CouchNotFoundError:
            # this means document is not exist so we will just insert
            pass
        finally:
            res = self.db.commitOne(doc)
        return res

    def user_update(self, request, msg, mtype='comment'):
        """Update user info in LogDB for given request"""
        rec = {"ts":tstamp(), "msg":msg}
        doc = {"_id": self.docid(request, mtype), "messages": [rec],
                "request":request, "identifier":self.dbid,
                "thr":self.thread_name, "type":mtype}
        try:
            exist_doc = self.db.document(doc["_id"])
            doc["_rev"] = exist_doc["_rev"]
            doc["messages"] += exist_doc["messages"]
        except CouchNotFoundError:
            # this means document is not exist so we will just insert
            pass
        finally:
            res = self.db.commitOne(doc)
        return res

    def get(self, request, mtype=None, detail=True, agent=True):
        """Retrieve all entries from LogDB for given request"""
        self.check(request, mtype)
        if agent and mtype:
            mtype = self.prefix(mtype)
        options = {'reduce':False}
        if mtype:
            keys = [[request, mtype]]
        else:
            keys=[]
            options.update({'startkey': [request], 'endkey':[request, {}]})
        if detail:
            options.update({'include_docs': True})
        docs = self.db.loadView(self.design, self.view, options, keys=keys)
        return docs

    def get_by_thread(self, request, mtype='error', detail=False, agent=True):
        self.check(request, mtype)
        if agent and mtype:
            mtype = self.prefix(mtype)
        keys = [[request, self.dbid, self.thread_name, mtype]]
        options = {'reduce':False}
        if detail:
            options.update({'include_docs': True})
        docs = self.db.loadView(self.design, self.threadview, options, keys)
        return docs

    def get_by_request(self, request):
        keys = [request]
        options = {'reduce':False}
        docs = self.db.loadView(self.design, self.requestview, options, keys)
        return docs

    def get_all_requests(self):
        """Retrieve all entries from LogDB"""
        options = {'reduce':True, 'group_level':1}
        docs = self.db.loadView(self.design, self.view, options)
        return docs

    def delete(self, request, mtype=None, this_thread=False, agent=True):
        """Delete entry in LogDB for given request"""
        if mtype:
            self.check(request, mtype)
        else:
            self.check(request)
        if this_thread:
            docs = self.get_by_thread(request, mtype=mtype, detail=False, agent=agent)
        else:
            docs = self.get(request, mtype=mtype, detail=False, agent=agent)
        ids = [r['id'] for r in docs.get('rows', [])]
        res = self.db.bulkDeleteByIDs(ids)
        return res

    def cleanup(self, thr):
        """
        Clean-up docs older then given threshold (thr should be specified in seconds).
        This is done via tstamp view end endkey, e.g.
        curl "http://127.0.0.1:5984/logdb/_design/LogDB/_view/tstamp?endkey=1427912282"
        """
        cutoff = int(round(time.time()-thr))
        #docs = self.db.allDocs() # may need another view to look-up old docs
        spec = {'endkey':cutoff, 'reduce':False}
        docs = self.db.loadView(self.design, self.tsview, spec)
        ids = [d['id'] for d in docs.get('rows', [])]
        self.db.bulkDeleteByIDs(ids)
        return ids



#!/usr/bin/env python
"""
_StdBase_

Base class with helper functions for standard WMSpec files.
"""
from __future__ import division
from future.utils import viewitems
from builtins import range, object

import logging
import json

from Utils.PythonVersion import PY3
from Utils.Utilities import decodeBytesToUnicodeConditional
from Utils.Utilities import makeList, makeNonEmptyList, strToBool, safeStr
from WMCore.WMRuntime.Tools.Scram import isCMSSWSupported
from WMCore.Cache.WMConfigCache import ConfigCache, ConfigCacheException
from WMCore.Lexicon import (couchurl, procstring, activity, procversion, primdataset,
                            gpuParameters, lfnBase, identifier, acqname, cmsname,
                            dataset, block, campaign, subRequestType)
from WMCore.ReqMgr.DataStructs.RequestStatus import REQUEST_START_STATE
from WMCore.ReqMgr.Tools.cms import releases, architectures
from WMCore.Services.Rucio.RucioUtils import RUCIO_RULES_PRIORITY
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.WMSpec.WMWorkloadTools import (makeLumiList, checkDBSURL, validateArgumentsCreate)


class StdBase(object):
    """
    _StdBase_

    Base class with helper functions for standard WMSpec file.
    """

    def __init__(self):
        """
        __init__

        Setup parameters that will be used by all workflows.  These parameters
        are not required to be set when the workflow is first created but will
        need to be set before the sandbox is created and jobs are made.
        These parameters in the init are set all to None, then they wil
        be overwritten in the call.

        These parameters can be changed after the workflow has been created by
        the methods in the WMWorkloadHelper class.
        """
        argumentDefinition = self.getWorkloadCreateArgs()
        for arg in argumentDefinition:
            setattr(self, argumentDefinition[arg]["attr"], None)

        # Internal parameters
        self.workloadName = None
        self.config_cache = {}

        return

    def __call__(self, workloadName, arguments):
        """
        __call__

        Look through the arguments that were passed into the workload's call
        method and pull out any that are setup by this base class.
        """
        self.workloadName = workloadName
        argumentDefinition = self.getWorkloadCreateArgs()
        for arg in argumentDefinition:
            try:
                if arg in arguments:
                    if arguments[arg] is None:
                        setattr(self, argumentDefinition[arg]["attr"], arguments[arg])
                    else:
                        value = arguments[arg]
                        setattr(self, argumentDefinition[arg]["attr"], value)
                elif argumentDefinition[arg]["optional"]:
                    defaultValue = argumentDefinition[arg]["default"]
                    setattr(self, argumentDefinition[arg]["attr"], defaultValue)
            except Exception as ex:
                raise WMSpecFactoryException("parameter %s: %s" % (arg, str(ex)))

        # TODO: this replace can be removed in one year from now, thus March 2022
        if hasattr(self, "dbsUrl"):
            self.dbsUrl = self.dbsUrl.replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")
            self.dbsUrl = self.dbsUrl.rstrip("/")

        return

    # static copy of the skim mapping
    skimMap = {}

    @staticmethod
    def calcEvtsPerJobLumi(ePerJob, ePerLumi, tPerEvent, requestedEvents=None):
        """
        _calcEvtsPerJobLumi_

        Given RequestNumEvents (for MC from scratch), EventsPerJob,
        EventsPerLumi and TimePerEvent information, calculates the final
        values for EventsPerJob and EventsPerLumi.

        Final result will always be an EventsPerJob multiple of EventsPerLumi,
        no matter whether EventsPerJob was provided or not. In addition to that,
        makes sure EventsPerJob is not greater than RequestNumEvents.
        :param ePerJob: events per job
        :param ePerLumi: events per lumi
        :param tPerEvent: time per event
        """
        # if not set, let's calculate an 8h job and set it for you
        if ePerJob is None:
            ePerJob = int((8.0 * 3600.0) / tPerEvent)
        if requestedEvents and ePerJob > requestedEvents:
            ePerJob = requestedEvents

        if ePerLumi is None:
            ePerLumi = ePerJob
        elif ePerLumi > ePerJob:
            ePerLumi = ePerJob
        else:
            # then make EventsPerJob multiple of EventsPerLumi and still closer to 8h jobs
            multiplier = int(round(ePerJob / ePerLumi))
            # make sure not to have 0 EventsPerJob
            multiplier = max(multiplier, 1)
            ePerJob = ePerLumi * multiplier

        return ePerJob, ePerLumi

    @staticmethod
    def skimToDataTier(cmsswVersion, skim):
        """
        Start subprocess and call CMSSW python code to retrieve data tier for given skim

        Detects a usable scram arch for this CMSSW release on this machine

        Cache the results and use cache for lookup if possible

        """
        if not cmsswVersion:
            return None

        if cmsswVersion in StdBase.skimMap:
            if skim in StdBase.skimMap[cmsswVersion]:
                return StdBase.skimMap[cmsswVersion][skim]
        else:
            StdBase.skimMap[cmsswVersion] = {}

        import glob
        import subprocess

        p = subprocess.Popen("/cvmfs/cms.cern.ch/common/cmsos", stdout=subprocess.PIPE, shell=True)
        cmsos = p.communicate()[0]
        cmsos = decodeBytesToUnicodeConditional(cmsos, condition=PY3).strip()

        scramBaseDirs = glob.glob("/cvmfs/cms.cern.ch/%s*/cms/cmssw/%s" % (cmsos, cmsswVersion))
        if not scramBaseDirs:
            scramBaseDirs = glob.glob("/cvmfs/cms.cern.ch/%s*/cms/cmssw-patch/%s" % (cmsos, cmsswVersion))
            if not scramBaseDirs:
                return None

        command = "source /cvmfs/cms.cern.ch/cmsset_default.sh\n"
        command += "cd %s\n" % scramBaseDirs[0]
        command += "eval `scramv1 runtime -sh`\n"

        if isCMSSWSupported(cmsswVersion, "CMSSW_10_3_0"):
            command += """python3 -c 'from Configuration.StandardSequences.Skims_cff import getSkimDataTier\n"""
        else:
            command += """python -c 'from Configuration.StandardSequences.Skims_cff import getSkimDataTier\n"""

        command += """dataTier = getSkimDataTier("%s")\n""" % skim
        command += """if dataTier:\n\tprint(dataTier.value())'"""

        p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
        dataTier = p.communicate()[0]
        dataTier = decodeBytesToUnicodeConditional(dataTier, condition=PY3).strip()

        if dataTier == "None":
            dataTier = None

        StdBase.skimMap[cmsswVersion][skim] = dataTier

        return dataTier

    def loadCouchID(self, configDoc=None, configCacheUrl=None, couchDBName=None):
        """
        Load a config document from couch db and return the object
        :param configDoc: the config ID document
        :param configCacheUrl: couch url for the config
        :param couchDBName: couch database name
        :return: config document object
        """
        # TODO: Evaluate if we should call validateConfigCacheExists here
        configCache = None
        if configDoc is not None and configDoc != "":
            if (configCacheUrl, couchDBName) in self.config_cache:
                configCache = self.config_cache[(configCacheUrl, couchDBName)]
            else:
                configCache = ConfigCache(configCacheUrl, couchDBName, True)
                self.config_cache[(configCacheUrl, couchDBName)] = configCache

            configCache.loadByID(configDoc)

        return configCache

    def determineOutputModules(self, scenarioFunc=None, scenarioArgs=None,
                               configCache=None, cmsswVersion=None):
        """
        _determineOutputModules_

        Determine the output module names and associated metadata for the
        given config.
        """
        # set default scenarioArgs to empty dictionary if it is None.
        scenarioArgs = scenarioArgs or {}
        outputModules = {}
        if configCache is not None:
            outputModules = configCache.getOutputModuleInfo()
        else:
            if 'outputs' in scenarioArgs and scenarioFunc in ["promptReco", "expressProcessing", "repack"]:

                for output in scenarioArgs.get('outputs', []):

                    moduleLabel = output['moduleLabel']
                    outputModules[moduleLabel] = {'dataTier': output['dataTier']}
                    if 'primaryDataset' in output:
                        outputModules[moduleLabel]['primaryDataset'] = output['primaryDataset']
                    if 'filterName' in output:
                        outputModules[moduleLabel]['filterName'] = output['filterName']

                for physicsSkim in scenarioArgs.get('PhysicsSkims', []):
                    dataTier = StdBase.skimToDataTier(cmsswVersion, physicsSkim)
                    if dataTier:
                        moduleLabel = "SKIMStream%s" % physicsSkim
                        outputModules[moduleLabel] = {'dataTier': dataTier,
                                                      'filterName': physicsSkim}
                    else:
                        self.raiseValidationException("Can't find physics skim %s in %s" % (physicsSkim, cmsswVersion))

            elif scenarioFunc == "alcaSkim":

                for alcaSkim in scenarioArgs.get('skims', []):
                    moduleLabel = "ALCARECOStream%s" % alcaSkim
                    if alcaSkim.startswith("PromptCalibProd"):
                        dataTier = "ALCAPROMPT"
                    else:
                        dataTier = "ALCARECO"
                    outputModules[moduleLabel] = {'dataTier': dataTier,
                                                  'primaryDataset': scenarioArgs.get('primaryDataset'),
                                                  'filterName': alcaSkim}

        return outputModules

    def addRuntimeMonitors(self, task):
        """
        _addRuntimeMonitors_

        Add dashboard monitoring for the given task.
        Memory settings are defined in Megabytes and timing in seconds.
        """
        # Default settings defined by CMS policy
        maxpss = 2.3 * 1024  # 2.3 GiB, but in MiB
        softTimeout = 47 * 3600  # 47h
        hardTimeout = 47 * 3600 + 5 * 60  # 47h + 5 minutes

        monitoring = task.data.section_("watchdog")
        monitoring.interval = 300
        monitoring.monitors = ["PerformanceMonitor"]
        monitoring.section_("PerformanceMonitor")
        monitoring.PerformanceMonitor.maxPSS = maxpss
        monitoring.PerformanceMonitor.softTimeout = softTimeout
        monitoring.PerformanceMonitor.hardTimeout = hardTimeout
        return task

    def createWorkload(self):
        """
        _createWorkload_

        Create a new workload.
        """
        ownerProps = {'dn': self.owner_dn, 'vogroup': self.owner_vogroup, 'vorole': self.owner_vorole}

        workload = newWorkload(self.workloadName)
        workload.setOwnerDetails(self.owner, self.group, ownerProps)
        workload.setStartPolicy("DatasetBlock", SliceType="NumberOfFiles", SliceSize=1)
        workload.setEndPolicy("SingleShot")
        workload.setAcquisitionEra(acquisitionEras=self.acquisitionEra)
        workload.setProcessingVersion(processingVersions=self.processingVersion)
        workload.setProcessingString(processingStrings=self.processingString)
        workload.setValidStatus(validStatus=self.validStatus)
        workload.setLumiList(lumiLists=self.lumiList)
        workload.setPriority(self.priority)
        workload.setCampaign(self.campaign)
        workload.setRequestType(self.requestType)
        workload.setDbsUrl(self.dbsUrl)
        workload.setPrepID(self.prepID)
        return workload

    def setupProcessingTask(self, procTask, taskType, inputDataset=None, inputStep=None,
                            inputModule=None, scenarioName=None,
                            scenarioFunc=None, scenarioArgs=None,
                            couchDBName=None, configDoc=None, splitAlgo="LumiBased",
                            splitArgs=None, seeding=None,
                            totalEvents=None, eventsPerLumi=None,
                            userDN=None, asyncDest=None, owner_vogroup="DEFAULT",
                            owner_vorole="DEFAULT", stepType="CMSSW",
                            userSandbox=None, userFiles=None, primarySubType=None,
                            forceMerged=False, forceUnmerged=False,
                            configCacheUrl=None, timePerEvent=None, memoryReq=None,
                            sizePerEvent=None, applySiteLists=True, cmsswVersion=None,
                            scramArch=None, globalTag=None, taskConf=None):
        """
        _setupProcessingTask_

        Given an empty task add cmsRun, stagOut and logArch steps.  Configure
        the input depending on the method parameters:
          inputDataset not empty: This is a top level processing task where the
            input will be fed in by DBS.  Setup the whitelists and blacklists.
          inputDataset empty: This processing task will be fed from the output
            of another task.  The inputStep and name of the output module from
            that step (inputModule) must be specified.

        Processing config will be setup as follows:
          configDoc not empty - Use a ConfigCache config, configCacheUrl and
            couchDBName must not be empty.
          configDoc empty - Use a Configuration.DataProcessing config.  The
            scenarioName, scenarioFunc and scenarioArgs parameters must not be
            empty.

        The seeding and totalEvents parameters are only used for production jobs.

        taskConf is a dictionary with either Task or Step level key/value pairs used to
        bypass the workload level argument, providing flexibility to have different
        settings even inside the same task object.
        """
        # set default values in case it's not passed to this method
        scenarioArgs = scenarioArgs or {}
        taskConf = taskConf or {}
        splitArgs = splitArgs or {'lumis_per_job': 8}
        userFiles = userFiles or []

        self.addRuntimeMonitors(procTask)
        procTaskCmssw = procTask.makeStep("cmsRun1")
        procTaskCmssw.setStepType(stepType)
        procTaskStageOut = procTaskCmssw.addStep("stageOut1")
        procTaskStageOut.setStepType("StageOut")
        procTaskStageOut.setUserDN(userDN)
        procTaskStageOut.setAsyncDest(asyncDest)
        procTaskStageOut.setUserRoleAndGroup(owner_vogroup, owner_vorole)
        procTaskStageOut.setNewStageoutOverride(self.enableNewStageout)
        procTaskLogArch = procTaskCmssw.addStep("logArch1")
        procTaskLogArch.setStepType("LogArchive")
        procTaskLogArch.setNewStageoutOverride(self.enableNewStageout)

        procTask.applyTemplates()

        procTask.setTaskLogBaseLFN(self.unmergedLFNBase)

        newSplitArgs = {}
        for argName in splitArgs:
            newSplitArgs[str(argName)] = splitArgs[argName]

        procTask.setSplittingAlgorithm(splitAlgo, **newSplitArgs)

        if not timePerEvent and self.timePerEvent:
            timePerEvent = self.timePerEvent
        if not sizePerEvent and self.sizePerEvent:
            sizePerEvent = self.sizePerEvent
        if not memoryReq and self.memory:
            memoryReq = self.memory
        if not cmsswVersion and self.frameworkVersion:
            cmsswVersion = self.frameworkVersion
        if not scramArch and self.scramArch:
            scramArch = self.scramArch
        if not globalTag and self.globalTag:
            globalTag = self.globalTag

        procTask.setJobResourceInformation(timePerEvent=timePerEvent,
                                           sizePerEvent=sizePerEvent,
                                           memoryReq=memoryReq)

        procTask.setTaskType(taskType)

        # we better be safe in case the specs set these Task/Step level args to None
        acqEra = taskConf.get("AcquisitionEra") or self.acquisitionEra
        procStr = taskConf.get("ProcessingString") or self.processingString
        procVer = taskConf.get("ProcessingVersion") or self.processingVersion
        prepID = taskConf.get("PrepID") or self.prepID
        campaignName = taskConf.get("Campaign") or self.campaign
        procTask.setAcquisitionEra(acqEra)
        procTask.setProcessingString(procStr)
        procTask.setProcessingVersion(procVer)
        procTask.setCampaignName(campaignName)

        if taskType in ["Production", 'PrivateMC'] and totalEvents is not None:
            procTask.addGenerator(seeding)
            procTask.addProduction(totalEvents=totalEvents)
            procTask.setFirstEventAndLumi(firstEvent=self.firstEvent,
                                          firstLumi=self.firstLumi)
        else:
            if inputDataset is not None:
                (primary, processed, tier) = self.inputDataset[1:].split("/")
                procTask.addInputDataset(name=self.inputDataset, primary=primary,
                                         processed=processed, tier=tier, dbsurl=self.dbsUrl,
                                         block_blacklist=self.blockBlacklist,
                                         block_whitelist=self.blockWhitelist,
                                         run_blacklist=self.runBlacklist,
                                         run_whitelist=self.runWhitelist)
            elif inputStep is not None and inputModule is not None:
                procTask.setInputReference(inputStep, outputModule=inputModule)

        if primarySubType:
            procTask.setPrimarySubType(subType=primarySubType)

        procTaskCmsswHelper = procTaskCmssw.getTypeHelper()
        procTaskStageHelper = procTaskStageOut.getTypeHelper()

        # StepChain overrides
        multicore = self.multicore
        eventStreams = self.eventStreams
        if 'Multicore' in taskConf and taskConf['Multicore'] > 0:
            multicore = taskConf['Multicore']
        if taskConf.get("EventStreams") is not None and taskConf['EventStreams'] >= 0:
            eventStreams = taskConf['EventStreams']
        procTaskCmsswHelper.setNumberOfCores(multicore, eventStreams)

        gpuRequired = self.requiresGPU
        if taskConf.get('RequiresGPU', None):
            gpuRequired = taskConf['RequiresGPU']
        # Note that GPUParams has already been validated
        if "GPUParams" in taskConf and json.loads(taskConf['GPUParams']):
            gpuParams = json.loads(taskConf['GPUParams'])
        else:
            gpuParams = json.loads(self.gPUParams)
        procTaskCmsswHelper.setGPUSettings(gpuRequired, gpuParams)

        procTaskCmsswHelper.setUserSandbox(userSandbox)
        procTaskCmsswHelper.setUserFiles(userFiles)
        procTaskCmsswHelper.setGlobalTag(globalTag)
        procTaskCmsswHelper.setPrepId(prepID)
        procTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)
        procTaskCmsswHelper.setErrorDestinationStep(stepName=procTaskLogArch.name())

        if forceMerged:
            procTaskStageHelper.setMinMergeSize(0, 0)
        elif forceUnmerged and not isinstance(forceUnmerged, list):
            procTaskStageHelper.disableStraightToMerge()
        else:
            procTaskStageHelper.setMinMergeSize(self.minMergeSize, self.maxMergeEvents)
            if forceUnmerged and isinstance(forceUnmerged, list):
                procTaskStageHelper.disableStraightToMergeForOutputModules(forceUnmerged)

        procTaskCmsswHelper.cmsswSetup(cmsswVersion,
                                       softwareEnvironment="",
                                       scramArch=scramArch)

        if "events_per_lumi" in newSplitArgs:
            eventsPerLumi = newSplitArgs["events_per_lumi"]
        procTaskCmsswHelper.setEventsPerLumi(eventsPerLumi)
        configCache = self.loadCouchID(configDoc, configCacheUrl, couchDBName)
        procTaskCmsswHelper.setPhysicsType(configCache)
        configOutput = self.determineOutputModules(scenarioFunc, scenarioArgs,
                                                   configCache, cmsswVersion)
        outputModules = {}
        for outputModuleName in configOutput:
            outputModule = self.addOutputModule(procTask,
                                                outputModuleName,
                                                configOutput[outputModuleName].get('primaryDataset',
                                                                                   self.inputPrimaryDataset),
                                                configOutput[outputModuleName]['dataTier'],
                                                configOutput[outputModuleName].get('filterName', None),
                                                forceMerged=forceMerged, forceUnmerged=forceUnmerged, taskConf=taskConf)
            outputModules[outputModuleName] = outputModule

        if configDoc is not None and configDoc != "":
            procTaskCmsswHelper.setConfigCache(configCacheUrl, configDoc, couchDBName)
        else:
            # delete dataset information from scenarioArgs
            if 'outputs' in scenarioArgs:
                for output in scenarioArgs['outputs']:
                    if 'primaryDataset' in output:
                        del output['primaryDataset']
            if 'primaryDataset' in scenarioArgs:
                del scenarioArgs['primaryDataset']

            procTaskCmsswHelper.setDataProcessingConfig(scenarioName, scenarioFunc,
                                                        **scenarioArgs)
        # only in the very end, in order to get it in for the children tasks as well
        procTask.setPrepID(prepID)

        # has to be done in the very end such that child tasks are set too
        procTask.setPerformanceMonitor(softTimeout=taskConf.get("SoftTimeout", None),
                                       gracePeriod=taskConf.get("GracePeriod", None))

        # Set procTask physics task type
        procTask.setPhysicsTaskType()

        return outputModules

    def _getDictionaryParams(self, prop, key, default=None):
        """
        Support dictonary format for property definition.
        acquisitionEra, processingString, processingVersion
        """
        if isinstance(prop, dict):
            return prop.get(key, default)
        else:
            return prop

    def addOutputModule(self, parentTask, outputModuleName,
                        primaryDataset, dataTier, filterName,
                        stepName="cmsRun1", forceMerged=False,
                        forceUnmerged=False, taskConf=None):
        """
        _addOutputModule_

        Add an output module to the given processing task.

        taskConf is used for multi-step tasks where diff output processed
        dataset name is desired
        """
        taskConf = taskConf or {}
        haveFilterName = (filterName is not None and filterName != "")
        haveProcString = (self.processingString is not None and self.processingString != "")
        haveRunNumber = (self.runNumber is not None and self.runNumber > 0)

        taskName = parentTask.name()
        if self.requestType == "StepChain" and "StepName" in taskConf:
            taskName = taskConf["StepName"]
        acqEra = taskConf.get('AcquisitionEra') or self._getDictionaryParams(self.acquisitionEra, taskName)
        procString = taskConf.get('ProcessingString') or self._getDictionaryParams(self.processingString, taskName)
        procVersion = taskConf.get('ProcessingVersion') or self._getDictionaryParams(self.processingVersion,
                                                                                     taskName, 1)

        processedDataset = "%s-" % acqEra
        if haveFilterName:
            processedDataset += "%s-" % filterName
        if haveProcString:
            processedDataset += "%s-" % procString
        processedDataset += "v%i" % procVersion

        if haveProcString:
            processingLFN = "%s-v%i" % (procString, procVersion)
        else:
            processingLFN = "v%i" % procVersion

        if haveRunNumber:
            stringRunNumber = str(self.runNumber).zfill(9)
            runSections = [stringRunNumber[i:i + 3] for i in range(0, 9, 3)]
            runLFN = "/".join(runSections)

        unmergedLFN = "%s/%s/%s/%s" % (self.unmergedLFNBase,
                                       acqEra,
                                       primaryDataset, dataTier)
        mergedLFN = "%s/%s/%s/%s" % (self.mergedLFNBase,
                                     acqEra,
                                     primaryDataset, dataTier)

        if haveFilterName:
            unmergedLFN += "/%s-%s" % (filterName, processingLFN)
            mergedLFN += "/%s-%s" % (filterName, processingLFN)
        else:
            unmergedLFN += "/%s" % processingLFN
            mergedLFN += "/%s" % processingLFN

        if haveRunNumber:
            unmergedLFN += "/%s" % runLFN
            mergedLFN += "/%s" % runLFN

        lfnBase(unmergedLFN)
        lfnBase(mergedLFN)

        isTransient = True

        if forceMerged:
            unmergedLFN = mergedLFN
            isTransient = False
        elif forceUnmerged:
            mergedLFN = unmergedLFN

        cmsswStep = parentTask.getStep(stepName)
        cmsswStepHelper = cmsswStep.getTypeHelper()
        cmsswStepHelper.addOutputModule(outputModuleName,
                                        primaryDataset=primaryDataset,
                                        processedDataset=processedDataset,
                                        dataTier=dataTier,
                                        filterName=filterName,
                                        lfnBase=unmergedLFN,
                                        mergedLFNBase=mergedLFN,
                                        transient=isTransient)

        return {"primaryDataset": primaryDataset,
                "dataTier": dataTier,
                "processedDataset": processedDataset,
                "filterName": filterName}

    def addLogCollectTask(self, parentTask, taskName="LogCollect", filesPerJob=500,
                          cmsswVersion=None, scramArch=None):
        """
        _addLogCollectTask_

        Create a LogCollect task for log archives that are produced by the
        parent task.
        """
        cmsswVersion = cmsswVersion or self.frameworkVersion
        scramArch = scramArch or self.scramArch
        logCollectTask = parentTask.addTask(taskName)
        self.addRuntimeMonitors(logCollectTask)
        logCollectStep = logCollectTask.makeStep("logCollect1")
        logCollectStep.setStepType("LogCollect")
        logCollectStep.setNewStageoutOverride(self.enableNewStageout)
        logCollectTask.applyTemplates()
        logCollectTask.setSplittingAlgorithm("MinFileBased", files_per_job=filesPerJob)
        logCollectTask.setTaskType("LogCollect")

        parentTaskLogArch = parentTask.getStep("logArch1")
        logCollectTask.setInputReference(parentTaskLogArch, outputModule="logArchive")

        logCollectStepHelper = logCollectStep.getTypeHelper()
        logCollectStepHelper.cmsswSetup(cmsswVersion,
                                        softwareEnvironment="",
                                        scramArch=scramArch)

        return logCollectTask

    def addMergeTask(self, parentTask, parentTaskSplitting, parentOutputModuleName, parentStepName="cmsRun1",
                     doLogCollect=True, lfn_counter=0, forceTaskName=None, cmsswVersion=None, scramArch=None,
                     taskConf=None):
        """
        _addMergeTask_

        Create a merge task for files produced by the parent task.

        taskConf is used for multi-step Tasks where different merge settings are desired for
        the same parent production/processing task
        """
        cmsswVersion = cmsswVersion or self.frameworkVersion
        scramArch = scramArch or self.scramArch
        # StepChain use case, to avoid merge task names clashes
        forceTaskName = forceTaskName or parentTask.name()
        taskConf = taskConf or {}

        mergeTask = parentTask.addTask("%sMerge%s" % (forceTaskName, parentOutputModuleName))
        self.addRuntimeMonitors(mergeTask)
        mergeTaskCmssw = mergeTask.makeStep("cmsRun1")
        mergeTaskCmssw.setStepType("CMSSW")

        mergeTaskStageOut = mergeTaskCmssw.addStep("stageOut1")
        mergeTaskStageOut.setStepType("StageOut")
        mergeTaskStageOut.setNewStageoutOverride(self.enableNewStageout)

        mergeTaskLogArch = mergeTaskCmssw.addStep("logArch1")
        mergeTaskLogArch.setStepType("LogArchive")
        mergeTaskLogArch.setNewStageoutOverride(self.enableNewStageout)

        mergeTask.setTaskLogBaseLFN(self.unmergedLFNBase)
        # we better be safe in case the specs set these Task/Step level args to None
        acqEra = taskConf.get("AcquisitionEra") or self.acquisitionEra
        procStr = taskConf.get("ProcessingString") or self.processingString
        procVer = taskConf.get("ProcessingVersion") or self.processingVersion
        mergeTask.setAcquisitionEra(acqEra)
        mergeTask.setProcessingString(procStr)
        mergeTask.setProcessingVersion(procVer)

        if doLogCollect:
            taskNameLC = "%s%sMergeLogCollect" % (forceTaskName, parentOutputModuleName)
            self.addLogCollectTask(mergeTask, taskName=taskNameLC,
                                   cmsswVersion=cmsswVersion, scramArch=scramArch)

        mergeTask.setTaskType("Merge")
        mergeTask.applyTemplates()

        if parentTaskSplitting == "EventBased" and parentTask.taskType() != "Production":
            splitAlgo = "WMBSMergeBySize"
        else:
            splitAlgo = "ParentlessMergeBySize"

        parentTaskCmssw = parentTask.getStep(parentStepName)
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)

        dataTier = getattr(parentOutputModule, "dataTier")
        mergeTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)

        mergeTaskCmsswHelper = mergeTaskCmssw.getTypeHelper()
        mergeTaskStageHelper = mergeTaskStageOut.getTypeHelper()

        mergeTaskCmsswHelper.cmsswSetup(cmsswVersion,
                                        softwareEnvironment="",
                                        scramArch=scramArch)

        mergeTaskCmsswHelper.setErrorDestinationStep(stepName=mergeTaskLogArch.name())
        mergeTaskCmsswHelper.setGlobalTag(self.globalTag)
        mergeTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        if splitAlgo != "WMBSMergeBySize" and self.robustMerge:
            mergeTaskCmsswHelper.setSkipBadFiles(True)

        mergeTask.setSplittingAlgorithm(splitAlgo,
                                        max_merge_size=self.maxMergeSize,
                                        min_merge_size=self.minMergeSize,
                                        max_merge_events=self.maxMergeEvents,
                                        max_wait_time=self.maxWaitTime,
                                        initial_lfn_counter=lfn_counter)

        if dataTier == "DQMIO":
            mergeTaskCmsswHelper.setDataProcessingConfig("do_not_use", "merge",
                                                         newDQMIO=True)
        elif dataTier in ("NANOAOD", "NANOAODSIM"):
            mergeTaskCmsswHelper.setDataProcessingConfig("do_not_use", "merge",
                                                         mergeNANO=True)
        else:
            mergeTaskCmsswHelper.setDataProcessingConfig("do_not_use", "merge")

        mergeTaskStageHelper.setMinMergeSize(0, 0)

        self.addOutputModule(mergeTask, "Merged",
                             primaryDataset=getattr(parentOutputModule, "primaryDataset"),
                             dataTier=getattr(parentOutputModule, "dataTier"),
                             filterName=getattr(parentOutputModule, "filterName"),
                             forceMerged=True, taskConf=taskConf)

        self.addCleanupTask(parentTask, parentOutputModuleName, forceTaskName, dataTier)
        if self.enableHarvesting and getattr(parentOutputModule, "dataTier") in ["DQMIO", "DQM"]:
            self.addDQMHarvestTask(mergeTask, "Merged",
                                   uploadProxy=self.dqmUploadProxy,
                                   periodic_harvest_interval=self.periodicHarvestInterval,
                                   doLogCollect=doLogCollect,
                                   dqmHarvestUnit=self.dqmHarvestUnit,
                                   cmsswVersion=cmsswVersion, scramArch=scramArch)

        # only in the very end, in order to get it in for the children tasks as well
        prepID = taskConf.get("PrepID") or parentTask.getPrepID()
        mergeTask.setPrepID(prepID)

        return mergeTask

    def addCleanupTask(self, parentTask, parentOutputModuleName, forceTaskName=None, dataTier=''):
        """
        _addCleanupTask_

        Create a cleanup task to delete files produces by the parent task.
        """
        if forceTaskName is None:
            forceTaskName = parentTask.name()

        cleanupTask = parentTask.addTask("%sCleanupUnmerged%s" % (forceTaskName, parentOutputModuleName))
        self.addRuntimeMonitors(cleanupTask)
        cleanupTask.setTaskType("Cleanup")

        parentTaskCmssw = parentTask.getStep("cmsRun1")
        cleanupTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)
        cleanupTask.setSplittingAlgorithm("SiblingProcessingBased", files_per_job=50)

        cleanupStep = cleanupTask.makeStep("cleanupUnmerged%s" % parentOutputModuleName)
        cleanupStep.setStepType("DeleteFiles")
        cleanupTask.applyTemplates()

        # TODO: for StepChain, it will still use the Step1 PrepID. It has to be fixed
        cleanupTask.setPrepID(parentTask.getPrepID())

        return

    def addDQMHarvestTask(self, parentTask, parentOutputModuleName, uploadProxy=None,
                          periodic_harvest_interval=0, periodic_harvest_sibling=False,
                          parentStepName="cmsRun1", doLogCollect=True, dqmHarvestUnit="byRun",
                          cmsswVersion=None, scramArch=None):
        """
        _addDQMHarvestTask_

        Create a DQM harvest task to harvest the files produces by the parent task.
        """
        cmsswVersion = cmsswVersion or self.frameworkVersion
        scramArch = scramArch or self.scramArch

        if periodic_harvest_interval:
            harvestType = "Periodic"
        else:
            harvestType = "EndOfRun"

        harvestTask = parentTask.addTask("%s%sDQMHarvest%s" % (parentTask.name(),
                                                               harvestType,
                                                               parentOutputModuleName))
        self.addRuntimeMonitors(harvestTask)
        harvestTaskCmssw = harvestTask.makeStep("cmsRun1")
        harvestTaskCmssw.setStepType("CMSSW")

        harvestTaskUpload = harvestTaskCmssw.addStep("upload1")
        harvestTaskUpload.setStepType("DQMUpload")
        harvestTaskLogArch = harvestTaskCmssw.addStep("logArch1")
        harvestTaskLogArch.setStepType("LogArchive")

        harvestTask.setTaskLogBaseLFN(self.unmergedLFNBase)
        if doLogCollect:
            taskNameLC = "%s%s%sDQMHarvestLogCollect" % (parentTask.name(),
                                                         parentOutputModuleName,
                                                         harvestType)
            self.addLogCollectTask(harvestTask, taskName=taskNameLC,
                                   cmsswVersion=cmsswVersion, scramArch=scramArch)

        harvestTask.setTaskType("Harvesting")
        harvestTask.applyTemplates()

        harvestTaskCmsswHelper = harvestTaskCmssw.getTypeHelper()
        harvestTaskCmsswHelper.cmsswSetup(cmsswVersion,
                                          softwareEnvironment="",
                                          scramArch=scramArch)

        harvestTaskCmsswHelper.setErrorDestinationStep(stepName=harvestTaskLogArch.name())
        harvestTaskCmsswHelper.setGlobalTag(self.globalTag)
        harvestTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        harvestTaskCmsswHelper.setUserLFNBase("/")

        parentTaskCmssw = parentTask.getStep(parentStepName)
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)

        harvestTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName)

        harvestTask.setSplittingAlgorithm("Harvest",
                                          periodic_harvest_interval=periodic_harvest_interval,
                                          periodic_harvest_sibling=periodic_harvest_sibling,
                                          dqmHarvestUnit=dqmHarvestUnit)

        datasetName = "/%s/%s/%s" % (getattr(parentOutputModule, "primaryDataset"),
                                     getattr(parentOutputModule, "processedDataset"),
                                     getattr(parentOutputModule, "dataTier"))

        if self.dqmConfigCacheID is not None:
            harvestTaskCmsswHelper.setConfigCache(self.configCacheUrl, self.dqmConfigCacheID, self.couchDBName)
            harvestTaskCmsswHelper.setDatasetName(datasetName)
        else:
            scenarioArgs = {'globalTag': self.globalTag,
                            'datasetName': datasetName,
                            'runNumber': self.runNumber,
                            'dqmSeq': self.dqmSequences}
            if self.globalTagConnect:
                scenarioArgs['globalTagConnect'] = self.globalTagConnect
            if getattr(parentOutputModule, "dataTier") == "DQMIO":
                scenarioArgs['newDQMIO'] = True
            harvestTaskCmsswHelper.setDataProcessingConfig(self.procScenario,
                                                           "dqmHarvesting",
                                                           **scenarioArgs)

        harvestTaskUploadHelper = harvestTaskUpload.getTypeHelper()
        harvestTaskUploadHelper.setProxyFile(uploadProxy)
        harvestTaskUploadHelper.setServerURL(self.dqmUploadUrl)

        # if this was a Periodic harvesting add another for EndOfRun
        if periodic_harvest_interval:
            self.addDQMHarvestTask(parentTask=parentTask, parentOutputModuleName=parentOutputModuleName,
                                   uploadProxy=uploadProxy,
                                   periodic_harvest_interval=0, periodic_harvest_sibling=True,
                                   parentStepName=parentStepName, doLogCollect=doLogCollect,
                                   dqmHarvestUnit=dqmHarvestUnit,
                                   cmsswVersion=cmsswVersion, scramArch=scramArch)

        return

    def setupPileup(self, task, pileupConfig, stepName=None):
        """
        _setupPileup_

        Setup pileup for every CMSSW step in the task, unless a stepName
        is given - StepChain case - then only setup pileup for that specific
        step (cmsRun1, cmsRun2, etc).
        pileupConfig has the following data structure:
            {'mc': ['/mc_pd/procds/tier'], 'data': ['/data_pd/procds/tier']}
        """
        for puType, puList in viewitems(pileupConfig):
            task.setInputPileupDatasets(puList)

        if stepName:
            stepHelper = task.getStepHelper(stepName)
            stepHelper.setupPileup(pileupConfig, self.dbsUrl)
            return

        for stepName in task.listAllStepNames():
            step = task.getStep(stepName)
            if step.stepType() != "CMSSW":
                continue
            stepHelper = task.getStepHelper(stepName)
            stepHelper.setupPileup(pileupConfig, self.dbsUrl)

        return

    def validateSchema(self, schema):
        """
        _validateSchema_

        Validate the schema prior to building the workload
        This function should be overridden by individual specs

        If something breaks, raise a WMSpecFactoryException.  A message
        in that excpetion will be transferred to an HTTP Error later on.
        """
        pass

    def validateWorkload(self, workload):
        """
        _validateWorkload_

        Just in case you have something that you want to validate
        after the workload gets created, this is where you should
        put it.

        If something breaks, raise a WMSpecFactoryException.  A message
        in that exception will be transferred to an HTTP Error later on.
        """
        pass

    def factoryWorkloadConstruction(self, workloadName, arguments):
        """
        _factoryWorkloadConstruction_

        Master build for ReqMgr - builds the entire workload
        and also performs the proper validation.

        Named this way so that nobody else will try to use this name.
        """
        if arguments.get('RequestType') == 'Resubmission':
            self.validateSchema(schema=arguments)
        else:
            self.masterValidation(schema=arguments)
            self.validateSchema(schema=arguments)

        workload = self.__call__(workloadName=workloadName, arguments=arguments)
        self.validateWorkload(workload)

        return workload

    def masterValidation(self, schema):
        """
        _masterValidation_

        This is validation for global inputs that have to be implemented for
        multiple types of workflows in the exact same way.

        This uses programatically the definitions in getWorkloadCreateArgs
        for type-checking, existence, null tests and the specific validation functions.

        Any spec-specific extras are implemented in the overriden validateSchema
        """
        # Validate the arguments according to the workload arguments definition
        argumentDefinition = self.getWorkloadCreateArgs()
        try:
            validateArgumentsCreate(schema, argumentDefinition)
        except Exception as ex:
            self.raiseValidationException(str(ex))
        return

    def raiseValidationException(self, msg):
        """
        _raiseValidationException_

        Inbuilt method for raising exception so people don't have
        to import WMSpecFactoryException all over the place.
        """

        logging.error("About to raise exception %s", msg)
        raise WMSpecFactoryException(message=msg)

    def validateConfigCacheExists(self, configID, configCacheUrl, couchDBName,
                                  getOutputModules=True):
        """
        _validateConfigCacheExists_

        If we have a configCache, we should probably try and load it.
        """

        if configID == '' or configID == ' ':
            self.raiseValidationException(msg="ConfigCacheID is invalid and cannot be loaded")

        if (configCacheUrl, couchDBName) in self.config_cache:
            configCache = self.config_cache[(configCacheUrl, couchDBName)]
        else:
            configCache = ConfigCache(dbURL=configCacheUrl, couchDBName=couchDBName, detail=getOutputModules)
            self.config_cache[(configCacheUrl, couchDBName)] = configCache

        try:
            # if detail option is set return outputModules
            return configCache.validate(configID)
        except ConfigCacheException as ex:
            self.raiseValidationException(str(ex))

    @staticmethod
    def getWorkloadCreateArgs():
        """
        _getWorkloadCreateArgs_

        This represents the authoritative list of request arguments that are
        allowed by the current spec class during request creation
        The list is formatted as a 2-level dictionary, the keys in the first level
        are the identifiers for the arguments processed by the current spec.
        The second level dictionary contains the information about that argument for
        validation:
        - default: Gives a default value if not provided,
                   this default value usually is good enough for a standard workflow. If the argument is not optional
                   and a default value is provided, this is only meant for test purposes.
        - type: A function that verifies the type of the argument, it may also cast it into the appropiate python type.
                If the input is not compatible with the expected type, this method must throw an exception.
        - optional: This boolean value indicates if the value must be provided or not by user
                    or inherited class can overwrite with default value.
        - assign_optional: This boolean value indicates if the value must be provided when workflow is assigned if False.

        - validate: A function which validates the input after type casting,
                    it returns True if the input is valid, it can throw exceptions on invalid input.
        - attr: This represents the name of the attribute corresponding to the argument in the WMSpec object.
        - null: This indicates if the argument can have None as its value.

        If above is not specified, automatically set by following default value
        - default: None
        - type: str
        - optional: True
        - assign_optional: True
        - validate: None
        - attr: change first letter to lower case
        - null: False
        Example:
        {
            RequestPriority : {'default' : 0,
                               'type' : int,
                               'optional' : False,
                               'validate' : lambda x : x > 0,
                               'attr' : 'priority',
                               'null' : False}
        }
        This replaces the old syntax in the __call__ of:

        self.priority = arguments.get("RequestPriority", 0)
        """
        # if key is not specified (and argument is optional), then it's set to the default value
        arguments = {"RequestType": {"default": "StdBase", "optional": False},
                     "RequestString": {"optional": False, "null": False},
                     # default value will be AcquisitionEra except when AcquistionEra is dict
                     "Campaign": {"default": "", "optional": True, "validate": campaign},
                     "CMSSWVersion": {"validate": lambda x: x in releases(),
                                      "optional": False, "attr": "frameworkVersion"},
                     "ScramArch": {"validate": lambda x: all([y in architectures() for y in x]),
                                   "optional": False, "type": makeNonEmptyList},
                     "GlobalTag": {"optional": False, "null": False},
                     "ConfigCacheID": {"optional": False},

                     "RequestPriority": {"default": 8000, "type": int, "attr": "priority",
                                         "validate": lambda x: (x >= 0 and x < 1e6)},
                     "Group": {"default": "DATAOPS"},
                     "PrepID": {"default": None, "null": True},
                     "OpenRunningTimeout": {"default": 0, "type": int, "null": False,
                                            "validate": lambda x: x >= 0},
                     "GlobalTagConnect": {"null": True},
                     "LumiList": {"default": {}, "type": makeLumiList},
                     "DbsUrl": {"default": "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader",
                                "null": True, "validate": checkDBSURL},
                     "DashboardHost": {"default": "cms-jobmon.cern.ch"},
                     "DashboardPort": {"default": 8884, "type": int,
                                       "validate": lambda x: x > 0},
                     "TimePerEvent": {"default": 12.0, "type": float, "validate": lambda x: x > 0},
                     "SizePerEvent": {"default": 512.0, "type": float, "validate": lambda x: x > 0},
                     "PeriodicHarvestInterval": {"default": 0, "type": int, "validate": lambda x: x >= 0},
                     "DQMHarvestUnit": {"default": "byRun", "type": str, "attr": "dqmHarvestUnit"},
                     "DQMUploadProxy": {"null": True, "attr": "dqmUploadProxy"},
                     "DQMUploadUrl": {"default": "https://cmsweb.cern.ch/dqm/dev", "attr": "dqmUploadUrl"},
                     "DQMSequences": {"default": [], "type": makeList, "attr": "dqmSequences"},
                     "DQMConfigCacheID": {"null": True, "attr": "dqmConfigCacheID"},
                     "EnableHarvesting": {"default": False, "type": strToBool},
                     "EnableNewStageout": {"default": False, "type": strToBool},
                     "IncludeParents": {"default": False, "type": strToBool},
                     "ConfigCacheUrl": {"default": "https://cmsweb.cern.ch/couchdb", "validate": couchurl},
                     "VoGroup": {"default": "unknown", "attr": "owner_vogroup"},
                     "VoRole": {"default": "unknown", "attr": "owner_vorole"},
                     "ValidStatus": {"default": "PRODUCTION",
                                     "validate": lambda x: x in ("PRODUCTION", "VALID")},
                     "OverrideCatalog": {"null": True},
                     "RunNumber": {"default": 0, "type": int},
                     "RobustMerge": {"default": True, "type": strToBool},
                     "Comments": {"default": ""},
                     "SubRequestType": {"default": "", "validate": subRequestType},
                     "RequiresGPU": {"default": "forbidden",
                                     "validate": lambda x: x in ("forbidden", "optional", "required")},
                     "GPUParams": {"default": json.dumps(None), "validate": gpuParameters},


                     # FIXME (Alan on 27/Mar/017): maybe used by T0 during creation???
                     "MinMergeSize": {"default": 2 * 1024 * 1024 * 1024, "type": int,
                                      "validate": lambda x: x > 0},
                     "MaxMergeSize": {"default": 4 * 1024 * 1024 * 1024, "type": int,
                                      "validate": lambda x: x > 0},
                     "MaxWaitTime": {"default": 24 * 3600, "type": int,
                                     "validate": lambda x: x > 0},
                     "MaxMergeEvents": {"default": 100000000, "type": int,
                                        "validate": lambda x: x > 0},

                     # parameters that can be overwritten during assignment
                     "AcquisitionEra": {"validate": acqname, "optional": False},
                     "ProcessingString": {"validate": procstring, "optional": False},
                     "ProcessingVersion": {"default": 1, "type": int, "validate": procversion},
                     "Memory": {"default": 2300.0, "type": float, "validate": lambda x: x > 0},
                     "Multicore": {"default": 1, "type": int, "validate": lambda x: x > 0},
                     "EventStreams": {"type": int, "default": 0, "validate": lambda x: x >= 0, "null": True},
                     "MergedLFNBase": {"default": "/store/data"},
                     "UnmergedLFNBase": {"default": "/store/unmerged"},
                     "DeleteFromSource": {"default": False, "type": strToBool},
                     }

        # these arguments are internally set by ReqMgr2 and should not be provided by the user
        reqmgrArgs = {"Requestor": {"attr": "owner", "optional": False},
                      "RequestorDN": {"attr": "owner_dn", "optional": False, "null": False},
                      "RequestName": {"optional": False, "null": False, "validate": identifier},
                      "RequestStatus": {"optional": False, "validate": lambda x: x == REQUEST_START_STATE},
                      "RequestTransition": {"optional": False, "type": list},
                      "PriorityTransition": {"optional": False, "type": list},
                      "RequestDate": {"optional": False, "type": list},
                      "CouchURL": {"default": "https://cmsweb.cern.ch/couchdb", "validate": couchurl},
                      "CouchDBName": {"default": "reqmgr_config_cache", "type": str, "validate": identifier},
                      "CouchWorkloadDBName": {"default": "reqmgr_workload_cache", "validate": identifier}
                      }

        arguments.update(reqmgrArgs)
        # Set defaults for the argument specification
        StdBase.setDefaultArgumentsProperty(arguments)

        return arguments

    @staticmethod
    def getWorkloadAssignArgs():
        """
        _getWorkloadAssignArgs_

        This represents the authoritative list of request arguments that are
        allowed by the current spec class during request assignment.

        For more information on how these arguments are built, please have a look
        at the docstring for getWorkloadCreateArgs.
        """
        # if key is not specified (and argument is optional), then it's set to the default value
        arguments = {"RequestPriority": {"type": int, "attr": "priority",
                                         "validate": lambda x: (x >= 0 and x < 1e6)},
                     "RequestStatus": {"assign_optional": False, "validate": lambda x: x == 'assigned'},
                     "Team": {"default": "", "type": safeStr, "assign_optional": False,
                              "validate": lambda x: len(x) > 0},
                     "AcquisitionEra": {"validate": acqname, "assign_optional": True},
                     "ProcessingString": {"validate": procstring, "assign_optional": True},
                     "ProcessingVersion": {"type": int, "validate": procversion, "assign_optional": True},
                     "Memory": {"type": float, "validate": lambda x: x > 0},
                     "Multicore": {"type": int, "validate": lambda x: x > 0},
                     "EventStreams": {"null": True, "type": int, "validate": lambda x: x >= 0},

                     "SiteBlacklist": {"default": [], "type": makeList,
                                       "validate": lambda x: all([cmsname(y) for y in x])},
                     "SiteWhitelist": {"default": [], "type": makeNonEmptyList, "assign_optional": False,
                                       "validate": lambda x: all([cmsname(y) for y in x])},
                     # PhEDEx subscription information
                     "CustodialSites": {"default": [], "type": makeList, "assign_optional": True,
                                        "validate": lambda x: all([cmsname(y) for y in x])},
                     "NonCustodialSites": {"default": [], "type": makeList, "assign_optional": True,
                                           "validate": lambda x: all([cmsname(y) for y in x])},
                     "SubscriptionPriority": {"default": "Low", "assign_optional": True,
                                              "validate": lambda x: x.lower() in RUCIO_RULES_PRIORITY},
                     "DeleteFromSource": {"default": False, "type": strToBool},
                     # merge settings
                     "UnmergedLFNBase": {"assign_optional": True},
                     "MergedLFNBase": {"assign_optional": True},
                     "MinMergeSize": {"default": 2 * 1024 * 1024 * 1024, "type": int,
                                      "validate": lambda x: x > 0},
                     "MaxMergeSize": {"default": 4 * 1024 * 1024 * 1024, "type": int,
                                      "validate": lambda x: x > 0},
                     "MaxWaitTime": {"default": 24 * 3600, "type": int,
                                     "validate": lambda x: x > 0},
                     "MaxMergeEvents": {"default": 100000000, "type": int,
                                        "validate": lambda x: x > 0},
                     # data location management
                     "TrustSitelists": {"default": False, "type": strToBool},
                     "TrustPUSitelists": {"default": False, "type": strToBool},
                     "AllowOpportunistic": {"default": False, "type": strToBool},
                     # from assignment: performance monitoring data
                     "SoftTimeout": {"default": 129600, "type": int, "validate": lambda x: x > 0},
                     "GracePeriod": {"default": 300, "type": int, "validate": lambda x: x > 0},
                     "HardTimeout": {"default": 129600 + 300, "type": int, "validate": lambda x: x > 0},
                     # Block closing information
                     "BlockCloseMaxWaitTime": {"default": 66400, "type": int, "validate": lambda x: x > 0},
                     "BlockCloseMaxFiles": {"default": 500, "type": int, "validate": lambda x: x > 0},
                     "BlockCloseMaxEvents": {"default": 25000000, "type": int, "validate": lambda x: x > 0},
                     "BlockCloseMaxSize": {"default": 5000000000000, "type": int, "validate": lambda x: x > 0},
                     # dashboard activity
                     "Dashboard": {"default": "production", "type": str, "validate": activity},
                     # Override parameters for step (EOS log location, etc
                     # set to "" string or None for eos-lfn-prefix if you don't want to save the log in eos
                     "Override": {"default": {"eos-lfn-prefix": "root://eoscms.cern.ch//eos/cms/store/logs/prod/recent/PRODUCTION"},
                                  "type": dict},
                     # Rucio rule subscription lifetime (used in ContainerRules by T0)
                     "DatasetLifetime": {"default": 0, "type": int, "assign_optional": True, "validate": lambda x: x >= 0},
                     }
        # Set defaults for the argument specification
        StdBase.setDefaultArgumentsProperty(arguments)

        return arguments

    @staticmethod
    def getChainCreateArgs(firstTask=False, generator=False):
        """
        _getChainCreateArgs_

        This represents the authoritative list of request arguments that are
        allowed in each chain (Step/Task) of chained request, during request creation.
        Additional especific arguments must be defined inside each spec class.

        For more information on how these arguments are built, please have a look
        at the docstring for getWorkloadCreateArgs.
        """
        arguments = {'AcquisitionEra': {'optional': True, 'validate': acqname},
                     # Campaign at chain level has no meaning but for CompOps/McM bookkeeping
                     "Campaign": {"optional": True, "validate": campaign},
                     'BlockBlacklist': {'default': [], 'null': False, 'optional': True, 'type': makeList,
                                        'validate': lambda x: all([block(y) for y in x])},
                     'BlockWhitelist': {'default': [], 'null': False, 'optional': True, 'type': makeList,
                                        'validate': lambda x: all([block(y) for y in x])},
                     'CMSSWVersion': {'attr': 'frameworkVersion', 'null': True, 'optional': True,
                                      'validate': lambda x: x in releases()},
                     'ConfigCacheID': {'optional': False, 'type': str},
                     'DataPileup': {'default': None, 'null': False, 'optional': True, 'type': str,
                                    'validate': dataset},
                     # for task/step level, the default EventStreams value (0) comes from the top level
                     'EventStreams': {'null': True, 'type': int, 'validate': lambda x: x >= 0},
                     'EventsPerJob': {'null': True, 'type': int, 'validate': lambda x: x > 0},
                     'EventsPerLumi': {'default': None, 'null': True, 'optional': True, 'type': int,
                                       'validate': lambda x: x > 0},
                     'FilesPerJob': {'default': 1, 'null': False, 'optional': True, 'type': int,
                                     'validate': lambda x: x > 0},
                     'FilterEfficiency': {'default': 1.0, 'null': False, 'optional': True, 'type': float,
                                          'validate': lambda x: x > 0.0},
                     'GlobalTag': {'optional': False, 'type': str},
                     "IncludeParents": {"null": False, "default": False, "type": strToBool},
                     'InputDataset': {'null': False, 'optional': generator or not firstTask, 'validate': dataset},
                     'InputFromOutputModule': {'default': None, 'null': False, 'optional': firstTask},
                     'KeepOutput': {'default': True, 'null': False, 'optional': True, 'type': strToBool},
                     'LheInputFiles': {'default': False, 'null': False, 'optional': True, 'type': strToBool},
                     'LumiList': {'default': {}, 'type': makeLumiList},
                     'LumisPerJob': {'default': 8, 'null': False, 'optional': True, 'type': int,
                                     'validate': lambda x: x > 0},
                     'MCPileup': {'default': None, 'null': False, 'optional': True, 'type': str,
                                  'validate': dataset},
                     'Memory': {'default': None, 'null': True, 'type': float, 'validate': lambda x: x > 0},
                     'Multicore': {'default': 0, 'type': int, 'validate': lambda x: x > 0},
                     'PrepID': {'default': None, 'null': True, 'optional': True, 'type': str},
                     "RequiresGPU": {"default": "forbidden",
                                     "validate": lambda x: x in ("forbidden", "optional", "required")},
                     "GPUParams": {"default": json.dumps(None), "validate": gpuParameters},
                     'PrimaryDataset': {'null': True, 'validate': primdataset, 'attr': 'inputPrimaryDataset'},
                     'ProcessingString': {'optional': True, 'validate': procstring},
                     'ProcessingVersion': {'type': int, 'validate': procversion},
                     'RequestNumEvents': {'null': False, 'optional': not generator, 'type': int,
                                          'validate': lambda x: x > 0},
                     'RunBlacklist': {'default': [], 'null': False, 'optional': True, 'type': makeList,
                                      'validate': lambda x: all([isinstance(y, int) and int(y) > 0 for y in x])},
                     'RunWhitelist': {'default': [], 'null': False, 'optional': True, 'type': makeList,
                                      'validate': lambda x: all([isinstance(y, int) and int(y) > 0 for y in x])},
                     'ScramArch': {'null': True, 'optional': True, 'type': makeNonEmptyList,
                                   'validate': lambda x: all([y in architectures() for y in x])},
                     'Seeding': {'default': 'AutomaticSeeding', 'null': False, 'optional': True, 'type': str,
                                 'validate': lambda x: x in ["ReproducibleSeeding", "AutomaticSeeding"]},
                     'SplittingAlgo': {'default': 'EventAwareLumiBased', 'null': False, 'optional': True,
                                       'validate': lambda x: x in ["EventBased", "LumiBased",
                                                                   "EventAwareLumiBased", "FileBased"]}}

        # Set defaults for the argument specification
        StdBase.setDefaultArgumentsProperty(arguments)

        return arguments

    @staticmethod
    def setDefaultArgumentsProperty(arguments):
        for arg in arguments:
            arguments[arg].setdefault("default", None)
            arguments[arg].setdefault("type", str)
            arguments[arg].setdefault("optional", True)
            arguments[arg].setdefault("assign_optional", True)
            arguments[arg].setdefault("null", False)
            arguments[arg].setdefault("validate", None)
            arguments[arg].setdefault("attr", arg[:1].lower() + arg[1:])
        return

    @classmethod
    def getTestArguments(cls):
        """
        _getTestArguments_

        Using the getWorkloadCreateArgs definition, build a request schema
        that may pass basic validation and create successfully a workload
        of the current spec. Only for testing purposes! Any use of this function
        outside of unit tests and integration tests may put your life in danger.
        Note that in some cases like ConfigCacheID, there is no default that will work
        and tests should specifically provide one.
        """
        workloadDefinition = cls.getWorkloadCreateArgs()
        schema = {}
        for arg in workloadDefinition:
            # Dashboard parameter must be re-defined for test purposes
            if arg == "DashboardHost":
                schema[arg] = "127.0.0.1"
            elif arg == "ConfigCacheUrl":
                from os import environ
                schema[arg] = environ["COUCHURL"]
            elif arg == "RequestDate":
                from time import gmtime
                schema[arg] = gmtime()[:6]
            elif arg == "RequestTransition":
                from time import time
                schema[arg] = [{"Status": REQUEST_START_STATE, "UpdateTime": int(time()), "DN": "Fake_DN"}]
            elif arg == "PriorityTransition":
                from time import time
                schema[arg] = [{"Priority": REQUEST_START_STATE, "UpdateTime": int(time()), "DN": "Fake_DN"}]
            elif arg == "RequestStatus":
                schema[arg] = REQUEST_START_STATE
            elif arg == "CouchDBName":
                schema[arg] = "reqmgr_config_cache_t"
            elif arg == "ScramArch":
                schema[arg] = "slc6_amd64_gcc491"
            elif arg == "SiteWhitelist":
                schema[arg] = ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC']
            elif arg == "GlobalTag":
                schema[arg] = "GT_DP_V1"
            elif arg == "InputDataset":
                schema[arg] = "/MinimumBias/ComissioningHI-v1/RAW"
            elif not workloadDefinition[arg]["optional"]:
                if workloadDefinition[arg]["type"] == str:
                    if arg == "InputDataset":
                        schema[arg] = "/MinimumBias/ComissioningHI-v1/RAW"
                    elif arg == "CMSSWVersion":
                        schema[arg] = "CMSSW_7_6_2"
                    elif arg == "RequestType":
                        schema[arg] = workloadDefinition[arg]['default']
                    else:
                        schema[arg] = "FAKE"
                elif workloadDefinition[arg]["type"] == int or workloadDefinition[arg]["type"] == float:
                    schema[arg] = 1
                else:  # A non-optional list or similar. Just copy.
                    schema[arg] = workloadDefinition[arg]['default']
            else:
                schema[arg] = workloadDefinition[arg]['default']

        taskStep = {'ConfigCacheID': 'FAKE',
                    'PrimaryDataset': 'FAKE',
                    'EventsPerJob': 100,
                    'GlobalTag': 'GT_DP_V1',
                    'RequestNumEvents': 1000000,
                    'Seeding': 'AutomaticSeeding',
                    'SplittingAlgo': 'EventBased'}
        if schema['RequestType'] == 'TaskChain':
            schema['Task1'] = taskStep
            schema['Task1'].update({'TaskName': 'Task1Name_Test',
                                    'TimePerEvent': 123})
        elif schema['RequestType'] == 'StepChain':
            schema['Step1'] = taskStep
            schema['Step1'].update({'StepName': 'Step1Name_Test'})

        return schema

    @classmethod
    def getAssignTestArguments(cls):
        """
        _getAssignTestArguments_

        Using the getWorkloadAssignArgs definition, build a request schema
        that may pass basic validation and successfully assign a workload
        of the current spec.

        Only for testing purposes! Any use of this function outside of unit
        tests and integration tests may put your life in danger.
        """
        workloadDefinition = cls.getWorkloadAssignArgs()
        schema = {}
        for arg in workloadDefinition:
            if arg == "SiteWhitelist":
                schema[arg] = ["T2_US_TEST_Site"]
            elif arg == "RequestStatus":
                schema[arg] = 'assigned'
            elif not workloadDefinition[arg]["assign_optional"]:
                if workloadDefinition[arg]["type"] == str:
                    if arg == "Team":
                        schema[arg] = "Test-Team"
                    else:
                        schema[arg] = "FAKE"
                elif workloadDefinition[arg]["type"] in (int, float):
                    schema[arg] = 1
                else:
                    schema[arg] = workloadDefinition[arg]['default']
            else:
                schema[arg] = workloadDefinition[arg]['default']
        return schema

    @staticmethod
    def validateGPUSettings(schemaData):
        """
        Method to check whether GPU settings have been provided for
        a workflow (or tasks/step) that requires GPUs (or has it set
        to optional).
        :param schemaData: workflow or task/step dictionary
        :return: nothing if validation is successful, otherwise raises an exception
        """
        if schemaData.get("RequiresGPU") in ("optional", "required"):
            try:
                msg = "Request is set with RequiresGPU={}, ".format(schemaData["RequiresGPU"])
                if not json.loads(schemaData["GPUParams"]):
                    msg += "but GPUParams argument is empty and/or incorrect."
                    raise WMSpecFactoryException(msg)
            except KeyError:
                msg += "but GPUParams argument has not been provided."
                raise WMSpecFactoryException(msg)
        return True



#!/usr/bin/env python
"""
_ListThresholdsForCreate_

Query the database to determine how many jobs are pending so that we can
determine whether or not to task for more work.
"""

from future.utils import viewitems

from WMCore.Database.DBFormatter import DBFormatter


class ListThresholdsForCreate(DBFormatter):
    assignedSQL = """SELECT wmbs_location.site_name, wmbs_location.pending_slots,
                            wmbs_location.cms_name, wmbs_location.plugin,
                            COUNT(wmbs_job.id) AS total, wls.name AS state,
                            runjob.status AS job_status, wmbs_workflow.priority
                            FROM wmbs_job
                       INNER JOIN wmbs_job_state ON
                         wmbs_job.state = wmbs_job_state.id
                       INNER JOIN wmbs_location ON
                         wmbs_job.location = wmbs_location.id
                       INNER JOIN wmbs_location_state wls ON
                         wls.id = wmbs_location.state
                       INNER JOIN wmbs_jobgroup ON
                         wmbs_jobgroup.id = wmbs_job.jobgroup
                       INNER JOIN wmbs_subscription ON
                         wmbs_subscription.id = wmbs_jobgroup.subscription
                       INNER JOIN wmbs_workflow ON
                         wmbs_workflow.id = wmbs_subscription.workflow
                       LEFT OUTER JOIN (SELECT wmbs_id AS id, bl_status.name as status
                                        FROM bl_runjob
                                            INNER JOIN bl_status ON
                                                bl_status.id = bl_runjob.sched_status
                                        WHERE bl_runjob.status = '1') runjob
                       ON runjob.id = wmbs_job.id
                     WHERE wmbs_job_state.name != 'success' AND
                           wmbs_job_state.name != 'complete' AND
                           wmbs_job_state.name != 'exhausted' AND
                           wmbs_job_state.name != 'cleanout' AND
                           wmbs_job_state.name != 'killed'
                     GROUP BY wmbs_location.site_name,
                     wmbs_location.pending_slots, wmbs_location.cms_name,
                     wls.name, runjob.status, wmbs_location.plugin,
                     wmbs_workflow.priority"""

    unassignedSQL = """SELECT wl.site_name, wl.pending_slots,
                              wl.cms_name, wls.name AS state,
                              COUNT(unassigned_jobs.job) AS job_count, unassigned_jobs.priority
                              FROM wmbs_location wl
                         LEFT OUTER JOIN
                           (SELECT DISTINCT wmbs_job_assoc.job, wmbs_location_pnns.location,
                                            wmbs_workflow.priority
                              FROM wmbs_job_assoc
                              INNER JOIN wmbs_file_location wfl ON
                                wmbs_job_assoc.fileid = wfl.fileid
                              INNER JOIN wmbs_location_pnns ON
                                wmbs_location_pnns.pnn = wfl.pnn
                              INNER JOIN wmbs_job ON
                                wmbs_job_assoc.job = wmbs_job.id
                              INNER JOIN wmbs_job_state ON
                                wmbs_job.state = wmbs_job_state.id
                              INNER JOIN wmbs_jobgroup ON
                                wmbs_job.jobgroup = wmbs_jobgroup.id
                              INNER JOIN wmbs_subscription ON
                                wmbs_jobgroup.subscription = wmbs_subscription.id
                              INNER JOIN wmbs_workflow ON
                                wmbs_workflow.id = wmbs_subscription.workflow
                              LEFT OUTER JOIN wmbs_subscription_validation wsv
                                ON wsv.location_id = wmbs_location_pnns.location
                                AND wsv.subscription_id = wmbs_jobgroup.subscription
                            WHERE wmbs_job.location IS NULL AND
                                  wmbs_job_state.name != 'killed' AND
                                  wmbs_job_state.name != 'cleanout' AND
                                  (wsv.valid = 1 OR
                                   (wsv.valid IS NULL AND NOT EXISTS
                                    (SELECT wsv2.valid FROM wmbs_subscription_validation wsv2
                                     WHERE wsv2.subscription_id = wmbs_jobgroup.subscription
                                     AND wsv2.valid = 1)))) unassigned_jobs ON
                            wl.id = unassigned_jobs.location
                         INNER JOIN wmbs_location_state wls ON
                            wls.id = wl.state
                            GROUP BY wl.site_name, wl.pending_slots,
                            wl.cms_name, wls.name, unassigned_jobs.priority"""

    def format(self, assignedResults, unassignedResults):
        """
        _format_

        Combine together the total we received from the assigned and unassigned
        queries into a single datastructure.
        """
        assignedResults = DBFormatter.formatDict(self, assignedResults)
        unassignedResults = DBFormatter.formatDict(self, unassignedResults)

        results = {}
        for result in assignedResults:
            if result["total"] == None:
                result["total"] = 0
            siteName = result["site_name"]

            if not siteName in results:
                results[siteName] = {"cms_name": result["cms_name"],
                                     "state": result["state"],
                                     "pending_jobs": {},
                                     "total_slots": result["pending_slots"]}

            countJobs = True
            if result['job_status']:
                module = __import__("WMCore.BossAir.Plugins.%s" % result['plugin'],
                                    globals(), locals(), [result['plugin']])
                plugIn = getattr(module, result['plugin'])
                status = plugIn.stateMap().get(result['job_status'])
                if status == 'Running':
                    countJobs = False

            priority = result["priority"]

            if priority not in results[siteName]["pending_jobs"]:
                results[siteName]["pending_jobs"][priority] = 0

            if countJobs:
                results[siteName]["pending_jobs"][priority] += result["total"]

            results[result["site_name"]]["total_slots"] = result["pending_slots"]

        # Sum up all the jobs currently unassigned
        for result in unassignedResults:
            siteName = result['site_name']
            if siteName not in results:
                results[siteName] = {"total_slots": result["pending_slots"],
                                     "pending_jobs": {},
                                     "cms_name": result["cms_name"],
                                     "state": result["state"]}
            priority = result["priority"] or 0

            if priority not in results[siteName]['pending_jobs']:
                results[siteName]['pending_jobs'][priority] = 0
            results[siteName]['pending_jobs'][priority] += result['job_count']

        return results

    def formatTable(self, formattedResults):
        """
        _formatTable_
        """
        results = []
        for k, v in viewitems(formattedResults):
            item = {}
            item['site'] = k
            item.update(v)
            results.append(item)
        return results

    def execute(self, conn=None, transaction=False, tableFormat=False):
        assignedResults = self.dbi.processData(self.assignedSQL, conn=conn,
                                               transaction=transaction)
        unassignedResults = self.dbi.processData(self.unassignedSQL, conn=conn,
                                                 transaction=transaction)
        results = self.format(assignedResults, unassignedResults)
        if tableFormat:
            return self.formatTable(results)
        else:
            return results



"""
__JobUpdaterPoller__

Poller module for the JobUpdater, takes care of the
actual updating work.

Created on Apr 16, 2013

@author: dballest
"""

import logging
import threading
from Utils.Timers import timeFunction
from WMCore.BossAir.BossAirAPI import BossAirAPI
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.WMException import WMException
from WMCore.Database.CMSCouch import CouchConflictError, CouchError
from WMCore.Database.CouchUtils import CouchConnectionError


class JobUpdaterException(WMException):
    """
    _JobUpdaterException_

    A job updater exception-handling class for the JobUpdaterPoller
    """
    pass


class JobUpdaterPoller(BaseWorkerThread):
    """
    _JobUpdaterPoller_

    Poller class for the JobUpdater
    """

    def __init__(self, config):
        """
        __init__
        """
        BaseWorkerThread.__init__(self)
        self.config = config

        self.bossAir = BossAirAPI(config=self.config)
        self.reqmgr2 = ReqMgr(self.config.General.ReqMgr2ServiceURL)
        self.workqueue = WorkQueue(self.config.WorkQueueManager.couchurl,
                                   self.config.WorkQueueManager.dbname)

        myThread = threading.currentThread()

        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        self.listWorkflowsDAO = self.daoFactory(classname="Workflow.ListForJobUpdater")
        self.updateWorkflowPrioDAO = self.daoFactory(classname="Workflow.UpdatePriority")
        self.executingJobsDAO = self.daoFactory(classname="Jobs.GetNumberOfJobsForWorkflowStatus")

    def setup(self, parameters=None):
        """
        _setup_
        """
        pass

    def terminate(self, parameters=None):
        """
        _terminate_

        Terminate gracefully.
        """
        pass

    @timeFunction
    def algorithm(self, parameters=None):
        """
        _algorithm_
        """
        try:
            logging.info("Synchronizing priorities with ReqMgr...")
            self.synchronizeJobPriority()
            logging.info("Priorities were synchronized, wait until the next cycle")
        except CouchConnectionError as ex:
            msg = "Caught CouchConnectionError exception in JobUpdater\n"
            msg += "transactions postponed until the next polling cycle\n"
            msg += str(ex)
            logging.exception(msg)
        except CouchConflictError as ex:
            msg = "Caught CouchConflictError exception in JobUpdater\n"
            msg += "transactions postponed until the next polling cycle\n"
            msg += str(ex)
            logging.exception(msg)
        except CouchError as ex:
            if ex.status is None and ex.reason is None:
                msg = "Couch Server error occured. Mostly due to time out\n"
                msg += str(ex)
                logging.exception(msg)
            else:
                raise JobUpdaterException(str(ex))
        except Exception as ex:
            errorStr = str(ex)
            if 'Connection refused' in errorStr or "timed out" in errorStr:
                logging.warning("Failed to sync priorities. Trying in the next cycle")
            else:
                msg = "Caught unexpected exception in JobUpdater: %s\n" % errorStr
                logging.exception(msg)
                raise JobUpdaterException(msg)

    def getWorkflowPrio(self, wflowName):
        """
        Given a request name, it fetches data from ReqMgr2 and returns
        the request priority
        :param wflowName: string with the request name
        :return: integer with the request priority
        """
        result = None
        try:
            result = self.reqmgr2.getRequestByNames(wflowName)[0]
            result = int(result[wflowName]['RequestPriority'])
        except Exception as ex:
            logging.error("Couldn't retrieve the priority of request %s", wflowName)
            logging.error("Error: %s", str(ex))
        return result

    def synchronizeJobPriority(self):
        """
        _synchronizeJobPriority_

        Check WMBS and WorkQueue for active workflows and compare with the
        ReqMgr for priority changes. If a priority change occurs
        then update the job priority in the batch system and
        the elements in the local queue that have not been injected yet.
        """
        # Update the priority of workflows that are not in WMBS and just in local queue
        priorityCache = {}
        workflowsToUpdate = {}
        # this call returns workflows that only have Available elements in local queue
        workflowsToCheck = self.workqueue.getAvailableWorkflows()
        for workflow, workqueuePrio in workflowsToCheck:
            if workflow not in priorityCache:
                requestPrio = self.getWorkflowPrio(workflow)
                if requestPrio is None:
                    continue

                priorityCache[workflow] = requestPrio
                if workqueuePrio != priorityCache[workflow]:
                    logging.info("Updating workqueue elements priority for request: %s", workflow)
                    self.workqueue.updatePriority(workflow, priorityCache[workflow])

        # Check the workflows in WMBS
        workflowsToUpdateWMBS = {}
        workflowsToCheck = self.listWorkflowsDAO.execute()
        # this loop contains multiple entries for the same workflow
        for workflowEntry in workflowsToCheck:
            workflow = workflowEntry['name']
            wmbsPrio = int(workflowEntry['workflow_priority'])
            if workflow not in priorityCache:
                requestPrio = self.getWorkflowPrio(workflow)
                if requestPrio is None:
                    continue

                priorityCache[workflow] = requestPrio
                if wmbsPrio != priorityCache[workflow]:
                    logging.info("Updating workqueue elements priority for request: %s", workflow)
                    self.workqueue.updatePriority(workflow, priorityCache[workflow])

                    # if there are jobs in wmbs executing state, update their prio in condor
                    if self.executingJobsDAO.execute(workflow) > 0:
                        logging.info("Updating condor jobs priority for request: %s", workflow)
                        self.bossAir.updateJobInformation(workflow,
                                                          requestPriority=priorityCache[workflow])
                    workflowsToUpdateWMBS[workflow] = priorityCache[workflow]

        if workflowsToUpdateWMBS:
            logging.info("Updating WMBS workflow priority for %d workflows.", len(workflowsToUpdateWMBS))
            self.updateWorkflowPrioDAO.execute(workflowsToUpdateWMBS)



#!/usr/bin/env python
"""
__WebTools__

Core of the DMWM web stack. Provides a basic set of features including security,
performance optimisations (code minification, zipping and correct headers and
etags)
"""



#!/usr/bin/env python
"""
_GetOutputParentLFNs_

Oracle implementation of Jobs.GetOutputParentLFNs
"""

from WMCore.WMBS.MySQL.Jobs.GetOutputParentLFNs import GetOutputParentLFNs as GetOutputParentLFNsMySQL

class GetOutputParentLFNs(GetOutputParentLFNsMySQL):
    pass



#!/usr/bin/env python
"""
_ThresholdBySite_

Query WMBS, ResourceControl and BossAir to give information about the thresholds
and the number of running jobs.
"""

from WMCore.Database.DBFormatter import DBFormatter

class ThresholdBySite(DBFormatter):
    sql = """SELECT wmbs_location.site_name AS site_name,
                    wmbs_location.pending_slots,
                    wmbs_location.running_slots,
                    rc_threshold.max_slots,
                    rc_threshold.pending_slots AS task_pending_slots,
                    wmbs_sub_types.name AS task_type,
                    job_count.running_jobs AS task_running_jobs,
                    job_count.pending_jobs AS task_pending_jobs
                    FROM wmbs_location
               INNER JOIN rc_threshold ON
                 wmbs_location.id = rc_threshold.site_id
               INNER JOIN wmbs_sub_types ON
                 rc_threshold.sub_type_id = wmbs_sub_types.id
               LEFT OUTER JOIN
                 (SELECT wmbs_job.location AS location, wmbs_subscription.subtype AS subtype,
                         SUM(CASE WHEN bl_runjob.sched_status =
                                 (SELECT id FROM bl_status where name = 'Running')
                             THEN 1 ELSE 0 END) AS running_jobs,
                         SUM(CASE WHEN bl_runjob.sched_status =
                             (SELECT id FROM bl_status where name = 'Idle')
                             THEN 1 ELSE 0 END) AS pending_jobs
                         FROM wmbs_job
                    INNER JOIN wmbs_jobgroup ON
                      wmbs_job.jobgroup = wmbs_jobgroup.id
                    INNER JOIN wmbs_subscription ON
                      wmbs_jobgroup.subscription = wmbs_subscription.id
                    INNER JOIN wmbs_job_state ON
                      wmbs_job.state = wmbs_job_state.id
                    INNER JOIN bl_runjob ON
                      bl_runjob.wmbs_id = wmbs_job.id
                  WHERE wmbs_job_state.name = 'executing' AND
                        bl_runjob.status = '1'
                  GROUP BY wmbs_job.location, wmbs_subscription.subtype) job_count ON
                  wmbs_location.id = job_count.location AND
                  wmbs_sub_types.id = job_count.subtype
              WHERE wmbs_location.site_name = :site"""

    def execute(self, site, conn = None, transaction = False, tableFormat = False):
        results = self.dbi.processData(self.sql, {'site': site},
                                       conn = conn, transaction = transaction)
        return self.formatDict(results)



"""
_InsertWorker_

Oracle implementation of InsertWorker
"""

__all__ = []



from WMCore.Agent.Database.MySQL.InsertWorker import InsertWorker \
     as InsertWorkerMySQL

class InsertWorker(InsertWorkerMySQL):
    pass



#!/usr/bin/env python
"""
The WorkflowUpdater poller component.
Among the actions performed by this component, we can list:
* find active workflows in the agent
* filter those that require pileup dataset
* find out the current location for the pileup datasets
* get a list of blocks available and locked by WM
* match those blocks with the current pileup config json file. In other words,
  blocks that are no longer locked and/or available need to be removed from the
  json file.
* update this json in the workflow sandbox
"""

import logging
import threading

from Utils.CertTools import cert, ckey
from Utils.IteratorTools import flattenList
from Utils.Timers import timeFunction, CodeTimer
from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.Services.pycurl_manager import RequestHandler
from WMCore.WMException import WMException
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.DAOFactory import DAOFactory


class WorkflowUpdaterException(WMException):
    """
    Specific WorkflowUpdaterPoller exception handling.
    """


class WorkflowUpdaterPoller(BaseWorkerThread):
    """
    Poller that does the actual work for updating workflows.
    """

    def __init__(self, config):
        """
        Initialize WorkflowUpdaterPoller object
        :param config: a Configuration object with the component configuration
        """
        BaseWorkerThread.__init__(self)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listActiveWflows = self.daoFactory(classname="Workflow.GetUnfinishedWorkflows")

        # parse mandatory attributes from the configuration
        self.config = config
        self.rucioAcct = getattr(config.WorkflowUpdater, "rucioAccount")
        self.rucioUrl = getattr(config.WorkflowUpdater, "rucioUrl")
        self.rucioAuthUrl = getattr(config.WorkflowUpdater, "rucioAuthUrl")
        self.rucioCustomScope = getattr(config.WorkflowUpdater, "rucioCustomScope",
                                        "group.wmcore")
        self.msPileupUrl = getattr(config.WorkflowUpdater, "msPileupUrl")

        self.userCert = cert()
        self.userKey = ckey()
        self.rucio = Rucio(acct=self.rucioAcct,
                           hostUrl=self.rucioUrl,
                           authUrl=self.rucioAuthUrl)
                           # configDict={'logger': self.logger})

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Executed in every polling cycle. The actual logic of the component is:
          1. find active workflows in the agent
          2. check if those active workflows are using pileup data
        :param parameters: not really used. But keeping same signature as
            the one defined in the super class.
        :return: only what is returned by the decorator
        """
        logging.info("Running Workflow updater injector poller algorithm...")
        try:
            # retrieve list of workflows with unfinished Production
            # or Processing subscriptions
            wflowSpecs = self.listActiveWflows.execute()
            if not wflowSpecs:
                logging.info("Agent has no active workflows at the moment")
                return

            # figure out workflows that have pileup
            puWflows = self.findWflowsWithPileup(wflowSpecs)
            if not puWflows:
                logging.info("Agent has no active workflows with pileup at the moment")
                return
            # resolve unique active pileup dataset names
            uniqueActivePU = flattenList([item['pileup'] for item in puWflows])

            # otherwise, move on retrieving pileups
            msPileupList = self.getPileupDocs()

            # and resolve blocks in each container being used by workflows
            # considerations for 2024 are around 100 pileups each taking 2 seconds in Rucio
            with CodeTimer("Rucio block resolution", logger=logging):
                self.findRucioBlocks(uniqueActivePU, msPileupList)

            # TODO in future tickets: find the json files in the spec/sandbox area and tweak them

        except Exception as ex:
            msg = f"Caught unexpected exception in WorkflowUpdater. Details:\n{str(ex)}"
            logging.exception(msg)
            raise WorkflowUpdaterException(msg) from None

    def getPileupDocs(self):
        """
        Fetch all pileup documents from MSPileup and preprocess the data.

        Note that the 'blocks' field is for the moment just a placeholder,
        as it will be populated in a later stage,

        :return: a list of dictionaries in the following format:
          {"pileupName": string with pileup name,
           "customName": string with custom pileup name - if any,
           "rses": list of RSE names,
           "blocks": list of block names}
        """
        mgr = RequestHandler()
        headers = {'Content-Type': 'application/json'}
        data = mgr.getdata(self.msPileupUrl, params={}, headers=headers, verb='GET',
                           ckey=self.userKey, cert=self.userCert, encode=True, decode=True)
        if data and data.get("result", []):
            if "error" in data["result"][0]:
                msg = f"Failed to retrieve MSPileup documents. Error: {data}"
                raise WorkflowUpdaterException(msg)

        logging.info("A total of %d pileup documents have been retrieved.", len(data["result"]))
        pileupMapList = []
        for puItem in data["result"]:
            logging.info("Pileup: %s, custom name: %s, expected at: %s, but currently available at: %s",
                         puItem['pileupName'], puItem['customName'],
                         puItem['expectedRSEs'], puItem['currentRSEs'])
            thisPU = {"pileupName": puItem['pileupName'],
                      "customName": puItem['customName'],
                      "rses": puItem['currentRSEs'],
                      "blocks": []}
            pileupMapList.append(thisPU)
        return pileupMapList

    def findWflowsWithPileup(self, listSpecs):
        """
        Given a list of workflow names and their respective specs, load each
        one of them and filter out those that don't require any pileup dataset.
        :param listSpecs: a list of dictionary with workflow name and spec path
        :return: a list of dictionaries with workflow name, spec path and list
            of pileup datasets being used, e.g.:
            {"name": string with workflow name,
             "spec": string with spec path,
             "pileup": list of strings with pileup names}
        """
        wflowsWithPU = []
        for wflowSpec in listSpecs:
            try:
                workloadHelper = WMWorkloadHelper()
                workloadHelper.load(wflowSpec['spec'])
                pileupSpecs = workloadHelper.listPileupDatasets()
                if pileupSpecs:
                    wflowSpec['pileup'] = pileupSpecs.values()
                    logging.info("Workflow: %s requires pileup dataset(s): %s",
                                 wflowSpec['name'], wflowSpec['pileup'])
                    wflowsWithPU.append(wflowSpec)
                else:
                    logging.info("Workflow: %s does not require any pileup", wflowSpec['name'])
            except Exception as ex:
                msg = f"Failed to load spec file for: {wflowSpec['spec']}. Details: {str(ex)}"
                logging.error(msg)
        logging.info("There are %d pileup workflows out of %d active workflows.",
                     len(wflowsWithPU), len(listSpecs))
        return wflowsWithPU

    def findRucioBlocks(self, uniquePUList, msPileupList):
        """
        Given a list of unique pileup dataset names, list all of
        their blocks in Rucio. Note that if a pileup document contains
        a customName dataset, then we need to resolve the blocks for that
        instead.
        :param uniquePUList: a list with pileup names
        :param msPileupList: a list with dictionaries from MSPileup
        :return: update the msPileupList object in place, by populating
            the 'block' field with a list of block names
        """
        for pileupItem in msPileupList:
            if pileupItem["pileupName"] not in uniquePUList:
                # no active workflow requires this pileup
                continue

            if pileupItem["customName"]:
                logging.info("Fetching blocks for custom pileup container: %s", pileupItem["customName"])
                pileupItem["blocks"] = self.rucio.getBlocksInContainer(pileupItem["customName"],
                                                                       scope=self.rucioCustomScope)
            else:
                logging.info("Fetching blocks for pileup container: %s", pileupItem["pileupName"])
                pileupItem["blocks"] = self.rucio.getBlocksInContainer(pileupItem["pileupName"], scope='cms')



#!/usr/bin/env python
"""
_JobsByState_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.JobCountByState import JobCountByState \
 as JobCountByStateMySQL

class JobCountByState(JobCountByStateMySQL):
    sql = JobCountByStateMySQL.sql



#!/usr/bin/env python
"""
_UpdateAlgo_

MySQL implementation of DBSBuffer.UpdateAlgo
"""




from WMCore.Database.DBFormatter import DBFormatter

class UpdateAlgo(DBFormatter):
    sql = """UPDATE dbsbuffer_algo SET in_dbs = :in_dbs
             WHERE app_name = :app_name AND app_ver = :app_ver AND
                   app_fam = :app_fam AND pset_hash = :pset_hash"""

    sqlID = """UPDATE dbsbuffer_algo SET in_dbs = :in_dbs
               WHERE id = :algoID"""

    def execute(self, inDBS, algoID = None, appName = None, appVer = None,
                appFam = None, psetHash = None, conn = None,
                transaction = False):
        binds = {"in_dbs": inDBS}

        if algoID == None:
            binds["app_name"] = appName
            binds["app_ver"] = appVer
            binds["app_fam"] = appFam
            binds["pset_hash"] = psetHash
            self.dbi.processData(self.sql, binds, conn = conn,
                                 transaction = transaction)
        else:
            binds["algoID"] = algoID
            self.dbi.processData(self.sqlID, binds, conn = conn,
                                 transaction = transaction)

        return



#!/usr/bin/env python
"""
_LoadFromNameAndTask_

MySQL implementation of Workflow.LoadFromNameAndTask
"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class LoadFromNameAndTask(DBFormatter):
    sql = """SELECT wmbs_workflow.id, wmbs_workflow.spec, wmbs_workflow.name,
                    wmbs_users.cert_dn as dn, wmbs_users.owner as owner,
                    wmbs_users.grp as grp, wmbs_users.group_name as vogrp,
                    wmbs_users.role_name as vorole, wmbs_workflow.task,
                    wmbs_workflow.type, wmbs_workflow.priority
             FROM wmbs_workflow
             INNER JOIN wmbs_users ON
               wmbs_workflow.owner = wmbs_users.id
             WHERE wmbs_workflow.name = :workflow AND wmbs_workflow.task = :task"""

    def execute(self, workflow, task, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, {"workflow": workflow, "task": task},
                                      conn=conn, transaction=transaction)
        return self.formatDict(result)[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Script to be executed with the WMAgent environment, does you first need to execute:
source apps/wmagent/etc/profile.d/init.sh

Provided a workflow name in the command line, it will find all the
local workqueue elements and print a summary of work/data location
for the elements sitting in the Available status
"""


from builtins import next

import sys
import os

from WMCore.Configuration import loadConfigurationFile
from WMCore.WorkQueue.WorkQueueBackend import WorkQueueBackend
from WMCore.WorkQueue.DataStructs.WorkQueueElement import possibleSites


def printElementsSummary(reqName, elements, queueUrl):
    """
    Print the local couchdb situation based on the WQE status
    """
    print("Summary for %s and request %s" % (queueUrl, reqName))
    for elem in elements:
        if elem['Status'] != "Available":
            continue
        targetSites = possibleSites(elem)
        commonDataLoc = commonDataLocation(elem)
        print("  Element '%s' has the following site intersection: %s, with common data location: %s"
              % (elem.id, targetSites, commonDataLoc))
        printDataLocation(elem)

        if not targetSites and commonDataLoc:
            print("    this workflow has to be assigned to: %s" % commonDataLoc)
        if not targetSites and not commonDataLoc:
            print("    this workflow has to be assigned with AAA flags enabled according to input/PU location")


def printDataLocation(element):
    """
    Print all the input data and their current location
    """
    print("    LQE has primary AAA: %s and secondary AAA: %s" % (element['NoInputUpdate'], element['NoPileupUpdate']))
    print("    LQE has Inputs: %s" % element['Inputs'])
    print("    LQE has ParentData: %s" % element['ParentData'])
    print("    LQE has PileupData: %s" % element['PileupData'])

def commonDataLocation(element):
    """
    Make an intersection of all the data location
    :param element: workqueue element object
    :return: list with the common **data** location
    """
    commonLoc = set()
    if element['PileupData']:
        commonLoc = set(next(iter(element['PileupData'].values())))
    if element['Inputs']:
        if commonLoc:
            commonLoc = commonLoc & set(next(iter(element['Inputs'].values())))
        else:
            commonLoc = set(list(element['Inputs'].values())[0])
    if element['ParentData']:
        tempLoc = list(element['ParentData'].values())
        parentLoc = set(tempLoc[0])
        for temp in tempLoc:
            parentLoc = parentLoc & set(temp)
        commonLoc = commonLoc & parentLoc
    return commonLoc

def main():
    if 'WMAGENT_CONFIG' not in os.environ:
        os.environ['WMAGENT_CONFIG'] = '/data/srv/wmagent/current/config/wmagent/config.py'
    config = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])

    if len(sys.argv) != 2:
        print("You must provide a request name")
        sys.exit(1)
    reqName = sys.argv[1]

    localWQBackend = WorkQueueBackend(config.WorkQueueManager.couchurl, db_name="workqueue")
    localDocIDs = localWQBackend.getElements(WorkflowName=reqName)
    printElementsSummary(reqName, localDocIDs, localWQBackend.queueUrl)

    sys.exit(0)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
"""
_LoadFromID_

MySQL implementation of Workflow.LoadFromID
"""

from WMCore.Database.DBFormatter import DBFormatter

class LoadFromID(DBFormatter):
    sql = """SELECT wmbs_workflow.id, wmbs_workflow.spec, wmbs_workflow.name,
                    wmbs_users.cert_dn as dn, wmbs_users.owner as owner,
                    wmbs_users.grp as grp, wmbs_users.group_name as vogrp,
                    wmbs_users.role_name as vorole, wmbs_workflow.task,
                    wmbs_workflow.type, wmbs_workflow.priority
             FROM wmbs_workflow
             INNER JOIN wmbs_users ON
               wmbs_workflow.owner = wmbs_users.id
             WHERE wmbs_workflow.id = :workflow"""

    def execute(self, workflow = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"workflow": workflow},
                         conn = conn, transaction = transaction)
        return self.formatDict(result)[0]



#!/usr/bin/env python
"""
_ListThresholdsForCreate_

Oracle implementation of ResourceControl.ListThresholdsForCreate
"""

from WMCore.ResourceControl.MySQL.ListThresholdsForCreate \
     import ListThresholdsForCreate as MySQLListThresholdsForCreate

class ListThresholdsForCreate(MySQLListThresholdsForCreate):
    pass



#!/usr/bin/env python
"""
_SetOutcomeBulk_

Oracle implementation of Jobs.SetOutcomeBulk
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.SetOutcomeBulk import SetOutcomeBulk as MySQLSetOutcomeBulk

class SetOutcomeBulk(MySQLSetOutcomeBulk):
    pass



#!/usr/bin/env python
"""
_List_

Oracle implementation of ListLocations
"""




from WMCore.WMBS.MySQL.Locations.List import List as ListLocationsMySQL

class List(ListLocationsMySQL):
    sql = ListLocationsMySQL.sql



#!/usr/bin/env python
"""
_DeleteFileset_

MySQL implementation of DeleteCheck

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class DeleteCheck(DBFormatter):
    sql = """DELETE FROM wmbs_fileset WHERE id = :fileset AND
           NOT EXISTS (SELECT id FROM wmbs_subscription WHERE fileset = :fileset AND id != :subscription)"""

    def execute(self, fileid, subid, conn = None, transaction = False):
        self.dbi.processData(self.sql, {'fileset': fileid, 'subscription': subid},
                             conn = conn, transaction = transaction)
        return True #or raise



#!/usr/bin/env python
"""
_CountFiles_

Oracle implementation of DBSBuffer.CountFiles
"""




from WMComponent.DBS3Buffer.MySQL.CountFiles import CountFiles as MySQLCountFiles

class CountFiles(MySQLCountFiles):
    """
    _CountFiles_

    """
    pass



#!/bin/env python


"""
_DefaultRetryAlgo_

This is the default.  It's alarmingly simple.
"""

import logging

from WMComponent.RetryManager.PlugIns.RetryAlgoBase import RetryAlgoBase


class DefaultRetryAlgo(RetryAlgoBase):
    """
    _DefaultRetryAlgo_

    This is the simple 'wait a bit' cooloff algo
    """


    def isReady(self, job, cooloffType):
        """
        Actual function that does the work

        """

        # Get the cooloff time
        cooloffDict = self.getAlgoParam(job['jobType'])
        cooloffTime = cooloffDict.get(cooloffType.lower(), None)

        if not cooloffTime:
            logging.error('Unknown cooloffTime for type %s: passing' %(type))
            return False

        currentTime = self.timestamp()
        if currentTime - job['state_time'] > cooloffTime:
            return True
        else:
            return False



#!/usr/bin/env python
"""
_Step.Executor.AlcaHarvest_

Implementation of an Executor for a AlcaHarvest step

"""
from __future__ import print_function

import logging
import os
import shutil
import stat
import subprocess

from Utils.Utilities import rootUrlJoin
from WMCore.FwkJobReport.Report import Report
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMSpec.Steps.Executor import Executor
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure


class AlcaHarvest(Executor):
    """
    _AlcaHarvest_

    Execute a AlcaHarvest Step

    """

    def pre(self, emulator=None):
        """
        _pre_

        Pre execution checks

        """
        # Are we using an emulator?
        if emulator != None:
            return emulator.emulatePre(self.step)

        logging.info("Steps.Executors.%s.pre called", self.__class__.__name__)
        return None

    def execute(self, emulator=None):
        """
        _execute_

        """
        # Are we using emulators again?
        if emulator != None:
            return emulator.emulate(self.step, self.job)

        logging.info("Steps.Executors.%s.execute called", self.__class__.__name__)

        # Search through steps for analysis files
        for step in self.stepSpace.taskSpace.stepSpaces():
            if step == self.stepName:
                # Don't try to parse your own report; it's not there yet
                continue
            stepLocation = os.path.join(self.stepSpace.taskSpace.location, step)
            logging.info("Beginning report processing for step %s", step)
            reportLocation = os.path.join(stepLocation, 'Report.pkl')
            if not os.path.isfile(reportLocation):
                logging.error("Cannot find report for step %s in space %s", step, stepLocation)
                continue

            # First, get everything from a file and 'unpersist' it
            stepReport = Report()
            stepReport.unpersist(reportLocation, step)

            # Don't upload nor stage out files from bad steps.
            if not stepReport.stepSuccessful(step):
                continue

            # Pulling out the analysis files from each step
            analysisFiles = stepReport.getAnalysisFilesFromStep(step)

            # make sure all conditions from this job get the same uuid
            uuid = makeUUID()

            condFiles2copy = []
            lumiFiles2copy = []

            # Working on analysis files
            for analysisFile in analysisFiles:

                # deal with sqlite files
                if analysisFile.FileClass == "ALCA":

                    sqlitefile = analysisFile.fileName.replace('sqlite_file:', '', 1)

                    filenamePrefix = "Run%d@%s@%s" % (self.step.condition.runNumber,
                                                      analysisFile.inputtag, uuid)
                    filenameDB = filenamePrefix + ".db"
                    filenameTXT = filenamePrefix + ".txt"

                    shutil.copy2(os.path.join(stepLocation, sqlitefile), filenameDB)

                    textoutput = "prepMetaData %s\n" % analysisFile.prepMetaData
                    textoutput += "prodMetaData %s\n" % analysisFile.prodMetaData

                    with open(filenameTXT, "w") as fout:
                        fout.write(textoutput)

                    os.chmod(filenameDB,
                             stat.S_IREAD | stat.S_IWRITE | stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)
                    os.chmod(filenameTXT,
                             stat.S_IREAD | stat.S_IWRITE | stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)

                    condFiles2copy.append(filenameDB)
                    condFiles2copy.append(filenameTXT)

                # deal with text files containing lumi info
                elif analysisFile.FileClass == "ALCATXT":

                    shutil.copy2(os.path.join(stepLocation, analysisFile.fileName), analysisFile.fileName)
                    lumiFiles2copy.append(analysisFile.fileName)

            # copy conditions files out and fake the job report
            addedOutputFJR = False
            if self.step.condition.lfnbase:
                logging.info("Copy out conditions files to %s", self.step.condition.lfnbase)
                for file2copy in condFiles2copy:

                    logging.info("==> copy %s", file2copy)

                    targetLFN = os.path.join(self.step.condition.lfnbase, file2copy)
                    targetPFN = "root://eoscms//eos/cms%s" % targetLFN

                    command = "env XRD_WRITERECOVERY=0 xrdcp -s -f %s %s" % (file2copy, targetPFN)

                    p = subprocess.Popen(command, shell=True,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.STDOUT)
                    output = p.communicate()[0]
                    if p.returncode > 0:
                        msg = 'Failure during condition copy to EOS:\n'
                        msg += '   %s\n' % output
                        logging.error(msg)
                        raise WMExecutionFailure(60319, "AlcaHarvestFailure", msg)

                    # add fake output file to job report
                    addedOutputFJR = True
                    stepReport.addOutputFile(self.step.condition.outLabel,
                                             aFile={'lfn': targetLFN,
                                                    'pfn': targetPFN,
                                                    'module_label': self.step.condition.outLabel})

            # copy luminosity files out
            if self.step.luminosity.url:
                logging.info("Copy out luminosity files to %s", self.step.luminosity.url)
                for file2copy in lumiFiles2copy:

                    logging.info("==> copy %s", file2copy)

                    targetPFN = rootUrlJoin(self.step.luminosity.url, file2copy)
                    if not targetPFN:
                        msg = 'No valid URL for lumi copy:\n'
                        msg += '   %s\n' % self.step.luminosity.url
                        logging.error(msg)
                        raise WMExecutionFailure(60319, "AlcaHarvestFailure", msg)

                    command = "env XRD_WRITERECOVERY=0 xrdcp -s -f %s %s" % (file2copy, targetPFN)

                    p = subprocess.Popen(command, shell=True,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.STDOUT)
                    output = p.communicate()[0]
                    if p.returncode > 0:
                        msg = 'Failure during copy to EOS:\n'
                        msg += '   %s\n' % output
                        logging.error(msg)
                        raise WMExecutionFailure(60319, "AlcaHarvestFailure", msg)

            if not addedOutputFJR:
                # no conditions from AlcaHarvest is a valid result, can
                # happen if calibration algorithms produced no output
                # due to not enough statistics or other reasons
                #
                # add fake placeholder output file to job report
                logging.info("==> no sqlite files from AlcaHarvest job, creating placeholder file record")
                stepReport.addOutputFile(self.step.condition.outLabel,
                                         aFile={'lfn': "/no/output",
                                                'pfn': "/no/output",
                                                'module_label': self.step.condition.outLabel})

            # Am DONE with report
            # Persist it
            stepReport.persist(reportLocation)

        return

    def post(self, emulator=None):
        """
        _post_

        Post execution checkpointing

        """
        # Another emulator check
        if emulator is not None:
            return emulator.emulatePost(self.step)

        logging.info("Steps.Executors.%s.post called", self.__class__.__name__)
        return None



#!/usr/bin/env python
"""
_DeleteWorkflow_

MySQL implementation of DeleteWorkflow

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    """
    Create a workflow ready for subscriptions
    """
    sql = """delete from wmbs_workflow where id = :id"""

    def getBinds(self, id=None):
        return self.dbi.buildbinds(self.dbi.makelist(id), 'id')

    def execute(self, id = -1, conn = None, transaction = False):
        self.dbi.processData(self.sql, self.getBinds(id),
                     conn = conn, transaction = transaction)
        return True #or raise



#!/usr/bin/env python
"""
_DQMHarvest_

Workflow for harvest an input dataset.
"""
from WMCore.WMSpec.StdSpecs.DataProcessing import DataProcessing


class DQMHarvestWorkloadFactory(DataProcessing):
    """
    _DQMHarvestWorkloadFactory_

    """

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a DQMHarvest workload with the given parameters.
        """
        DataProcessing.__call__(self, workloadName, arguments)

        self.workload = self.createWorkload()

        self.workload.setDashboardActivity("harvesting")

        splitArgs = {"runs_per_job": 1}
        if self.dqmHarvestUnit == "multiRun":
            # then it should result in a single job in the end, very high number of runs
            splitArgs['runs_per_job'] = 999999
        self.workload.setWorkQueueSplitPolicy("Dataset", "Harvest", splitArgs)

        # also creates the logCollect job by default
        self.addDQMHarvestTask(uploadProxy=self.dqmUploadProxy,
                               periodic_harvest_interval=self.periodicHarvestInterval,
                               dqmHarvestUnit=self.dqmHarvestUnit)

        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        self.workload.setTaskPropertiesFromWorkload()

        return self.workload

    def validateSchema(self, schema):
        """
        _validateSchema_

        Standard DataProcessing schema validation.
        """
        DataProcessing.validateSchema(self, schema)

        self.validateConfigCacheExists(configID=schema["DQMConfigCacheID"],
                                       configCacheUrl=schema['ConfigCacheUrl'],
                                       couchDBName=schema["CouchDBName"],
                                       getOutputModules=False)

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = DataProcessing.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "DQMHarvest", "optional": True},
                    "ConfigCacheID": {"optional": True, "null": True},
                    "DQMConfigCacheID": {"optional": False, "attr": "dqmConfigCacheID"},
                    "DQMUploadUrl": {"optional": False, "attr": "dqmUploadUrl"},
                   }
        baseArgs.update(specArgs)
        DataProcessing.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    def addDQMHarvestTask(self, parentTask=None, parentOutputModuleName=None, uploadProxy=None,
                          periodic_harvest_interval=0, periodic_harvest_sibling=False,
                          parentStepName="cmsRun1", doLogCollect=True, dqmHarvestUnit="byRun",
                          cmsswVersion=None, scramArch=None):
        """
        _addDQMHarvestTask_

        Create a DQM harvest task that does not depend on previous tasks.
        """
        if periodic_harvest_interval:
            harvestType = "Periodic"
        else:
            harvestType = "EndOfRun"

        harvestTask = self.workload.newTask("%sDQMHarvest" % harvestType)

        self.addRuntimeMonitors(harvestTask)
        harvestTaskCmssw = harvestTask.makeStep("cmsRun1")
        harvestTaskCmssw.setStepType("CMSSW")

        harvestTaskUpload = harvestTaskCmssw.addStep("upload1")
        harvestTaskUpload.setStepType("DQMUpload")
        harvestTaskLogArch = harvestTaskCmssw.addStep("logArch1")
        harvestTaskLogArch.setStepType("LogArchive")

        harvestTask.setTaskLogBaseLFN(self.unmergedLFNBase)
        if doLogCollect:
            self.addLogCollectTask(harvestTask, taskName="%sDQMHarvestLogCollect" % harvestType)

        harvestTask.setTaskType("Harvesting")
        harvestTask.applyTemplates()

        harvestTaskCmsswHelper = harvestTaskCmssw.getTypeHelper()
        harvestTaskCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                          scramArch=self.scramArch)

        harvestTaskCmsswHelper.setErrorDestinationStep(stepName=harvestTaskLogArch.name())
        harvestTaskCmsswHelper.setGlobalTag(self.globalTag)
        harvestTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        harvestTaskCmsswHelper.setUserLFNBase("/")

        (inputPrimaryDataset, inputProcessedDataset, inputDataTier) = self.inputDataset[1:].split("/")
        harvestTask.addInputDataset(name=self.inputDataset,
                                    primary=inputPrimaryDataset,
                                    processed=inputProcessedDataset,
                                    tier=inputDataTier,
                                    dbsurl=self.dbsUrl,
                                    block_whitelist=self.blockWhitelist,
                                    block_blacklist=self.blockBlacklist,
                                    run_whitelist=self.runWhitelist,
                                    run_blacklist=self.runBlacklist)

        harvestTask.setSplittingAlgorithm("Harvest",
                                          periodic_harvest_interval=periodic_harvest_interval,
                                          periodic_harvest_sibling=periodic_harvest_sibling,
                                          dqmHarvestUnit=dqmHarvestUnit)

        harvestTask.setJobResourceInformation(memoryReq=self.memory)

        if self.dqmConfigCacheID is not None:
            harvestTaskCmsswHelper.setConfigCache(self.configCacheUrl, self.dqmConfigCacheID, self.couchDBName)
            harvestTaskCmsswHelper.setDatasetName(self.inputDataset)
        else:
            scenarioArgs = {'globalTag': self.globalTag,
                            'datasetName': self.inputDataset,
                            'runNumber': self.runNumber,
                            'dqmSeq': self.dqmSequences}
            if self.globalTagConnect:
                scenarioArgs['globalTagConnect'] = self.globalTagConnect
            if self.inputDataTier == "DQMIO":
                scenarioArgs['newDQMIO'] = True
            harvestTaskCmsswHelper.setDataProcessingConfig(self.procScenario,
                                                           "dqmHarvesting",
                                                           **scenarioArgs)

        harvestTaskUploadHelper = harvestTaskUpload.getTypeHelper()
        harvestTaskUploadHelper.setProxyFile(uploadProxy)
        harvestTaskUploadHelper.setServerURL(self.dqmUploadUrl)

        # if this was a Periodic harvesting add another for EndOfRun
        if periodic_harvest_interval:
            self.addDQMHarvestTask(uploadProxy=uploadProxy, periodic_harvest_interval=0,
                                   periodic_harvest_sibling=True, parentStepName=parentStepName,
                                   doLogCollect=doLogCollect, dqmHarvestUnit=dqmHarvestUnit)

        return
